# Copyright 2018-2026 Jérôme Dumonteil
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
from __future__ import annotations

from collections.abc import Iterable

import pytest

from odfdo.const import ODF_STYLES
from odfdo.document import Document
from odfdo.element import Element
from odfdo.page_layout import StylePageLayout
from odfdo.styles import Styles


@pytest.fixture
def styles(samples) -> Iterable[Styles]:
    document = Document(samples("example.odt"))
    yield document.get_part(ODF_STYLES)


def test_style_page_layout_create():
    spl = StylePageLayout()
    assert spl.serialize() == "<style:page-layout/>"


def test_style_page_layout_from_tag():
    spl = Element.from_tag("<style:page-layout/>")
    assert isinstance(spl, StylePageLayout)


def test_style_page_layout_repr():
    spl = Element.from_tag("<style:page-layout/>")
    assert repr(spl) == "<StylePageLayout family=page-layout name=None>"


def test_style_page_layout_create_name():
    spl = StylePageLayout(name="a_name")
    assert spl.serialize() == '<style:page-layout style:name="a_name"/>'


def test_style_page_layout_page_usage_1():
    spl = StylePageLayout(page_usage="bad")
    assert spl.serialize() == "<style:page-layout/>"


def test_style_page_layout_page_usage_2():
    spl = StylePageLayout(page_usage="all")
    assert spl.serialize() == "<style:page-layout/>"


def test_style_page_layout_page_usage_3():
    spl = StylePageLayout(page_usage="left")
    assert spl.serialize() == '<style:page-layout style:page-usage="left"/>'


def test_style_page_layout_page_usage_4():
    spl = StylePageLayout(page_usage="right")
    assert spl.serialize() == '<style:page-layout style:page-usage="right"/>'


def test_style_page_layout_page_usage_5():
    spl = StylePageLayout(page_usage="mirrored")
    assert spl.serialize() == '<style:page-layout style:page-usage="mirrored"/>'


def test_get_style_page_layout(styles):
    auto = styles.office_automatic_styles
    spl = auto.page_layouts[0]
    assert isinstance(spl, StylePageLayout)


def test_get_style_page_layout_name(styles):
    auto = styles.office_automatic_styles
    spl = auto.page_layouts[0]
    assert spl.name == "Mpm1"


def test_get_style_page_layout_family(styles):
    auto = styles.office_automatic_styles
    spl = auto.page_layouts[0]
    assert spl.family == "page-layout"


def test_get_style_page_layout_set_family_unchanged(styles):
    auto = styles.office_automatic_styles
    spl = auto.page_layouts[0]
    spl.family = "unchanged"
    assert spl.family == "page-layout"


def test_get_style_page_layout_page_usage(styles):
    auto = styles.office_automatic_styles
    spl = auto.page_layouts[0]
    usage = spl.page_usage
    assert usage == "all"


def test_get_style_page_layout_get_properties(styles):
    auto = styles.office_automatic_styles
    spl = auto.page_layouts[0]
    props = spl.get_properties()
    expected = {
        "fo:page-width": "20.999cm",
        "fo:page-height": "29.699cm",
        "style:num-format": "1",
        "style:print-orientation": "portrait",
        "fo:margin-top": "2cm",
        "fo:margin-bottom": "2cm",
        "fo:margin-left": "2cm",
        "fo:margin-right": "2cm",
        "style:writing-mode": "lr-tb",
        "style:layout-grid-color": "#c0c0c0",
        "style:layout-grid-lines": "20",
        "style:layout-grid-base-height": "0.706cm",
        "style:layout-grid-ruby-height": "0.353cm",
        "style:layout-grid-mode": "none",
        "style:layout-grid-ruby-below": "false",
        "style:layout-grid-print": "false",
        "style:layout-grid-display": "false",
        "style:footnote-max-height": "0cm",
        "style:footnote-sep": {
            "style:width": "0.018cm",
            "style:distance-before-sep": "0.101cm",
            "style:distance-after-sep": "0.101cm",
            "style:line-style": "none",
            "style:adjustment": "left",
            "style:rel-width": "25%",
            "style:color": "#000000",
        },
    }
    assert props == expected


def test_get_style_page_layout_set_properties(styles):
    auto = styles.office_automatic_styles
    spl = auto.page_layouts[0]
    new_props = {
        "fo:page-width": "10.999cm",
        "fo:page-height": "19.699cm",
        "style:num-format": "1",
        "style:print-orientation": "portrait",
        "fo:margin-top": "4cm",
        "fo:margin-bottom": "4cm",
        "fo:margin-left": "4cm",
        "fo:margin-right": "4cm",
        "style:writing-mode": "lr-tb",
        "style:layout-grid-color": "#cccccc",
        "style:layout-grid-lines": "10",
        "style:layout-grid-base-height": "0.706cm",
        "style:layout-grid-ruby-height": "0.353cm",
        "style:layout-grid-mode": "none",
        "style:layout-grid-ruby-below": "false",
        "style:layout-grid-print": "true",
        "style:layout-grid-display": "false",
        "style:footnote-max-height": "0cm",
        "style:footnote-sep": {
            "style:width": "0.018cm",
            "style:distance-before-sep": "0.101cm",
            "style:distance-after-sep": "0.101cm",
            "style:line-style": "none",
            "style:adjustment": "left",
            "style:rel-width": "25%",
            "style:color": "#000000",
        },
    }
    spl.set_properties(new_props)
    props = spl.get_properties()
    expected = new_props
    assert props == expected


def test_get_style_page_layout_del_properties(styles):
    auto = styles.office_automatic_styles
    spl = auto.page_layouts[0]
    expected = {
        "fo:page-width": "20.999cm",
        "fo:page-height": "29.699cm",
        "style:num-format": "1",
        "style:print-orientation": "portrait",
        "fo:margin-left": "2cm",
        "fo:margin-right": "2cm",
        "style:writing-mode": "lr-tb",
        "style:layout-grid-color": "#c0c0c0",
        "style:layout-grid-lines": "20",
        "style:layout-grid-base-height": "0.706cm",
        "style:layout-grid-ruby-height": "0.353cm",
        "style:layout-grid-mode": "none",
        "style:layout-grid-ruby-below": "false",
        "style:layout-grid-print": "false",
        "style:layout-grid-display": "false",
        "style:footnote-max-height": "0cm",
        "style:footnote-sep": {
            "style:width": "0.018cm",
            "style:distance-before-sep": "0.101cm",
            "style:distance-after-sep": "0.101cm",
            "style:line-style": "none",
            "style:adjustment": "left",
            "style:rel-width": "25%",
            "style:color": "#000000",
        },
    }
    spl.del_properties(["fo:margin-top", "fo:margin-bottom"])
    props = spl.get_properties()
    assert props == expected


def test_get_style_page_layout_set_background():
    spl = StylePageLayout(name="a_name")
    spl.set_background(color="#666666", position="center")
    assert spl.serialize() == (
        '<style:page-layout style:name="a_name">'
        '<style:page-layout-properties fo:background-color="#666666"/>'
        "</style:page-layout>"
    )


def test_get_style_page_layout_get_header_style(styles):
    auto = styles.office_automatic_styles
    spl = auto.page_layouts[0]
    style = spl.get_header_style()
    assert style.tag == "style:header-style"


def test_get_style_page_layout_set_header_style(styles):
    auto = styles.office_automatic_styles
    spl = auto.page_layouts[0]
    new_style = Element.from_tag("style:header-style")
    spl.set_header_style(new_style)
    style = spl.get_header_style()
    assert style.tag == "style:header-style"
    assert style.children == []


def test_get_style_page_layout_set_header_style_2(styles):
    auto = styles.office_automatic_styles
    spl = auto.page_layouts[0]
    current_style = spl.get_header_style()
    current_style.delete()
    new_style = Element.from_tag("style:header-style")
    spl.set_header_style(new_style)
    style = spl.get_header_style()
    assert style.tag == "style:header-style"
    assert style.children == []


def test_get_style_page_layout_get_footer_style(styles):
    auto = styles.office_automatic_styles
    spl = auto.page_layouts[0]
    style = spl.get_footer_style()
    assert style.tag == "style:footer-style"


def test_get_style_page_layout_set_footer_style(styles):
    auto = styles.office_automatic_styles
    spl = auto.page_layouts[0]
    new_style = Element.from_tag("style:footer-style")
    spl.set_footer_style(new_style)
    style = spl.get_footer_style()
    assert style.tag == "style:footer-style"
    assert style.children == []


def test_get_style_page_layout_set_footer_style_2(styles):
    auto = styles.office_automatic_styles
    spl = auto.page_layouts[0]
    current_style = spl.get_footer_style()
    current_style.delete()
    new_style = Element.from_tag("style:footer-style")
    spl.set_footer_style(new_style)
    style = spl.get_footer_style()
    assert style.tag == "style:footer-style"
    assert style.children == []



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python

from __future__ import annotations

from collections.abc import Iterable

import pytest

from odfdo import Element
from odfdo.form_controls import FormGrid


@pytest.fixture
def form_grid() -> Iterable[FormGrid]:
    yield FormGrid(
        name="some name",
        title="some title",
        control_implementation="some implem",
        disabled=False,
        tab_index=4,
        xml_id="control1",
    )


def test_form_grid_class():
    form = FormGrid()
    assert isinstance(form, FormGrid)


def test_form_grid_minimal_tag():
    form = Element.from_tag("<form:grid/>")
    assert isinstance(form, FormGrid)


def test_form_grid_serialize():
    form = FormGrid()
    assert form.serialize() == "<form:grid/>"


def test_form_grid_repr():
    form = FormGrid()
    assert repr(form) == "<FormGrid name=None xml_id=None>"


def test_form_grid_name(form_grid):
    assert form_grid.name == "some name"


def test_form_grid_title(form_grid):
    assert form_grid.title == "some title"


def test_form_grid_control_implementation(form_grid):
    assert form_grid.control_implementation == "some implem"


def test_form_grid_disabled(form_grid):
    assert form_grid.disabled is False


def test_form_grid_printable(form_grid):
    assert form_grid.printable is True


def test_form_grid_tab_index(form_grid):
    assert form_grid.tab_index == 4


def test_form_grid_tab_stop_1(form_grid):
    assert form_grid.tab_stop is True


def test_form_grid_tab_stop_2(form_grid):
    form_grid.tab_stop = False
    assert form_grid.tab_stop is False


def test_form_grid_xml_id(form_grid):
    assert form_grid.xml_id == "control1"


def test_form_grid_form_id(form_grid):
    assert form_grid.form_id == "control1"


def test_form_grid_xforms_bind():
    form = FormGrid(
        name="some name",
        value="some value",
        control_implementation="some implem",
        xml_id="control1",
        xforms_bind="bind id",
    )
    assert form.xforms_bind == "bind id"


def test_form_grid_form_id_2():
    form = FormGrid(
        name="some name",
        value="some value",
        control_implementation="some implem",
        xml_id="control1",
        form_id="control1",
    )
    assert form.form_id == "control1"



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2013 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
"""MetaHyperlinkBehaviour class for "meta:hyperlink-behaviour" tag."""

from __future__ import annotations

from typing import Any

from .element import Element, PropDef, PropDefBool, register_element_class


class MetaHyperlinkBehaviour(Element):
    """Container for hyperlink-behaviour properties, "meta:hyperlink-
    behaviour".
    """

    _tag = "meta:hyperlink-behaviour"
    _properties: tuple[PropDef | PropDefBool, ...] = (
        PropDef("target_frame_name", "office:target-frame-name"),
        PropDef("show", "xlink:show"),
    )

    def __init__(
        self,
        target_frame_name: str = "_blank",
        show: str = "replace",
        **kwargs: Any,
    ) -> None:
        """Initialize a MetaHyperlinkBehaviour element.

        The `meta:hyperlink-behaviour` element specifies the default behavior
        for hyperlinks in a document.

        Args:
            target_frame_name: The name of the target frame for the hyperlink.
                Defaults to "_blank" (new window/tab).
            show: Specifies how the target resource is presented.
                Defaults to "replace".
            **kwargs: Additional keyword arguments for the parent `Element` class.
        """
        super().__init__(**kwargs)

        if self._do_init:
            self.target_frame_name = target_frame_name
            self.show = show

    def __repr__(self) -> str:
        return (
            f"<{self.__class__.__name__} tag={self.tag} "
            f"target={self.target_frame_name} show={self.show}>"
        )

    def __str__(self) -> str:
        return f"({self.target_frame_name})"

    def as_dict(self) -> dict[str, Any]:
        """Return the attributes of the hyperlink behavior element as a Python dictionary.

        Returns:
            dict[str, Any]: A dictionary containing the hyperlink behavior
                attributes, with keys like "office:target-frame-name" and "xlink:show".
        """
        return {
            "office:target-frame-name": self.target_frame_name,
            "xlink:show": self.show,
        }

    def from_dict(self, data: dict[str, Any]) -> None:
        """Set the attributes of the hyperlink behavior element from a Python dictionary.

        Args:
            data: A dictionary containing the hyperlink behavior
                attributes (e.g., "office:target-frame-name", "xlink:show").
        """
        self.target_frame_name = str(data.get("office:target-frame-name", ""))
        self.show = str(data.get("xlink:show", "replace"))


MetaHyperlinkBehaviour._define_attribut_property()

register_element_class(MetaHyperlinkBehaviour)



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
# Authors: Hervé Cauwelier <herve@itaapy.com>

from __future__ import annotations

from collections.abc import Iterable

import pytest

from odfdo.document import Document
from odfdo.element import Element
from odfdo.frame import (
    Frame,
)
from odfdo.header import Header
from odfdo.image import DrawImage
from odfdo.paragraph import Paragraph

ZOE = "你好 Zoé"


@pytest.fixture
def frame_body(samples) -> Iterable[Element]:
    document = Document(samples("frame_image.odp"))
    yield document.body


def test_frame_minimal():
    frame = Frame()
    expected = (
        '<draw:frame draw:z-index="0" svg:height="1cm" svg:width="1cm"></draw:frame>'
    )
    assert frame._canonicalize() == expected


def test_frame_minimal_2():
    frame = Element.from_tag(
        '<draw:frame draw:z-index="0" svg:height="1cm" svg:width="1cm"></draw:frame>'
    )
    assert isinstance(frame, Frame)


def test_create_frame():
    frame = Frame("A Frame", size=("10cm", "12cm"), style="Graphics")
    expected = (
        '<draw:frame draw:name="A Frame" '
        'draw:style-name="Graphics" '
        'draw:z-index="0" '
        'svg:height="12cm" '
        'svg:width="10cm">'
        "</draw:frame>"
    )
    assert frame._canonicalize() == expected


def test_create_frame_ids():
    frame = Frame(
        "A Frame",
        draw_id="some id",
        presentation_class="some classes",
        presentation_style="presentation style",
        layer="some layer",
        size=("10cm", "12cm"),
        style="Graphics",
    )
    expected = (
        '<draw:frame draw:id="some id" '
        'draw:layer="some layer" '
        'draw:name="A Frame" '
        'draw:style-name="Graphics" '
        'draw:z-index="0" '
        'presentation:class="some classes" '
        'presentation:style-name="presentation style" '
        'svg:height="12cm" '
        'svg:width="10cm">'
        "</draw:frame>"
    )
    assert frame._canonicalize() == expected


def test_create_frame_page():
    frame = Frame(
        "Another Frame",
        size=("10cm", "12cm"),
        anchor_type="page",
        anchor_page=1,
        position=("14mm", "15mm"),
        style="Graphics",
    )
    expected = (
        "<draw:frame "
        'draw:name="Another Frame" '
        'draw:style-name="Graphics" '
        'draw:z-index="0" '
        'svg:height="12cm" '
        'svg:width="10cm" '
        'svg:x="14mm" '
        'svg:y="15mm" '
        'text:anchor-page-number="1" '
        'text:anchor-type="page">'
        "</draw:frame>"
    )
    assert frame._canonicalize() == expected


def test_get_frame_list(frame_body):
    result = frame_body.get_frames()
    assert len(result) == 4


def test_get_frame_list_property(frame_body):
    result = frame_body.frames
    assert len(result) == 4


def test_get_frame_list_title(frame_body):
    result = frame_body.get_frames(title="Intitulé")
    assert len(result) == 1
    assert result[0].tag == "draw:frame"


def test_get_frame_by_name(frame_body):
    frame = frame_body.get_frame(name="odfdo")
    assert frame.tag == "draw:frame"


def test_get_frame_by_position(frame_body):
    frame = frame_body.get_frame(position=3)
    assert frame.presentation_class == "notes"


def test_get_frame_by_description(frame_body):
    frame = frame_body.get_frame(description="描述")
    assert frame.tag == "draw:frame"


def test_insert_frame(frame_body):
    body = frame_body.clone
    frame1 = Frame("frame1", size=("10cm", "12cm"), style="Graphics")
    frame2 = Frame(
        "frame2",
        size=("10cm", "13cm"),
        anchor_page=1,
        position=("10mm", "10mm"),
        style="Graphics",
    )
    body.append(frame1)
    body.append(frame2)
    result = body.get_frames(style="Graphics")
    assert len(result) == 2
    element = body.get_frame(name="frame1")
    assert element.tag == "draw:frame"
    element = body.get_frame(name="frame2")
    assert element.tag == "draw:frame"


def test_create_image_frame():
    frame = Frame.image_frame("Pictures/zoe.jpg")
    expected = (
        "<draw:frame "
        'draw:z-index="0" '
        'svg:height="1cm" '
        'svg:width="1cm">'
        '<draw:image xlink:actuate="onLoad" '
        'xlink:href="Pictures/zoe.jpg" '
        'xlink:show="embed" '
        'xlink:type="simple">'
        "</draw:image>"
        "</draw:frame>"
    )
    assert frame._canonicalize() == expected


def test_image_frame_set_image_element_on_previous():
    frame = Frame.image_frame("Pictures/zoe.jpg")
    image = DrawImage("Pictures/new.jpg")
    frame.set_image(image)
    read_image = frame.get_image()
    assert read_image.url == "Pictures/new.jpg"


def test_image_frame_set_image_element_on_none():
    frame = Frame("frame name")
    image = DrawImage("Pictures/new.jpg")
    frame.set_image(image)
    read_image = frame.get_image()
    assert read_image.url == "Pictures/new.jpg"


def test_image_frame_set_image_url():
    frame = Frame.image_frame("Pictures/zoe.jpg")
    url = "Pictures/new.jpg"
    frame.set_image(url)
    read_image = frame.get_image()
    assert read_image.url == "Pictures/new.jpg"


def test_create_image_frame_text():
    frame = Frame.image_frame("Pictures/zoe.jpg", text=ZOE)
    expected = (
        "<draw:frame "
        'draw:z-index="0" '
        'svg:height="1cm" '
        'svg:width="1cm">'
        '<draw:image xlink:actuate="onLoad" '
        'xlink:href="Pictures/zoe.jpg" '
        'xlink:show="embed" '
        'xlink:type="simple">'
        f"<text:p>{ZOE}</text:p>"
        "</draw:image></draw:frame>"
    )
    assert frame._canonicalize() == expected


def test_create_text_frame():
    frame = Frame.text_frame(ZOE)
    expected = (
        '<draw:frame draw:z-index="0" '
        'svg:height="1cm" '
        'svg:width="1cm">'
        "<draw:text-box>"
        f"<text:p>{ZOE}</text:p>"
        "</draw:text-box>"
        "</draw:frame>"
    )
    assert frame._canonicalize() == expected


def test_frame_get_text_content():
    frame = Frame.text_frame(ZOE)
    expected = ZOE
    assert frame.text_content == expected


def test_frame_get_text_content_none():
    frame = Frame()
    assert frame.text_content == ""


def test_frame_set_text_content_1():
    frame = Frame()
    frame.text_content = "content"
    assert frame.text_content == "content"


def test_frame_set_text_content_2():
    frame = Frame.text_frame("previous")
    new_content = Paragraph("new")
    frame.text_content = new_content
    assert frame.text_content == "new"


def test_framset_text_box():
    frame = Frame.text_frame(ZOE)
    frame.set_text_box("some text")
    assert frame.text_content == "some text"


def test_frame_set_text_box_list():
    frame = Frame.text_frame(ZOE)
    frame.set_text_box(["some text", "second text"])
    assert frame.text_content == "some text\nsecond text"


def test_create_text_frame_element():
    heading = Header(1, ZOE)
    frame = Frame.text_frame(heading)
    expected = (
        '<draw:frame draw:z-index="0" '
        'svg:height="1cm" '
        'svg:width="1cm">'
        "<draw:text-box>"
        "<text:h "
        f'text:outline-level="1">{ZOE}</text:h>'
        "</draw:text-box>"
        "</draw:frame>"
    )
    assert frame._canonicalize() == expected


def test_get_frame(frame_body):
    body = frame_body
    frame = body.get_frame()
    assert isinstance(frame, Frame)


def test_get_frame_size():
    size = ("1cm", "2mm")
    position = ("3in", "4pt")
    frame = Frame(size=size, position=position, anchor_type="paragraph")
    assert frame.size == size


def test_set_size():
    size = ("1cm", "2mm")
    position = ("3in", "4pt")
    frame = Frame(size=size, position=position, anchor_type="paragraph")
    frame2 = frame.clone
    frame.size = position
    assert frame.size == position
    assert frame2.size == size


def test_get_frame_position():
    size = ("1cm", "2mm")
    position = ("3in", "4pt")
    frame = Frame(size=size, position=position, anchor_type="paragraph")
    assert frame.position == position


def test_set_frame_position():
    size = ("1cm", "2mm")
    position = ("3in", "4pt")
    frame = Frame(size=size, position=position, anchor_type="paragraph")
    frame2 = frame.clone
    frame2.position = size
    assert frame2.position == size


def test_get_frame_anchor_type():
    size = ("1cm", "2mm")
    position = ("3in", "4pt")
    frame = Frame(size=size, position=position, anchor_type="paragraph")
    assert frame.anchor_type == "paragraph"


def test_set_frame_anchor_type():
    size = ("1cm", "2mm")
    position = ("3in", "4pt")
    frame = Frame(size=size, position=position, anchor_type="paragraph")
    assert frame.anchor_page is None
    frame.anchor_type = "page"
    frame.anchor_page = 3
    assert frame.anchor_type == "page"
    assert frame.anchor_page == 3


def test_svg_description_1():
    frame = Element.from_tag(
        '<draw:frame svg:width="10cm" svg:height="10cm" '
        'draw:z-index="0" draw:name="Another Frame" '
        'draw:style-name="Graphics" '
        'svg:x="10mm" svg:y="10mm" '
        'text:anchor-type="page" '
        'text:anchor-page-number="1"/>'
    )
    result = frame.svg_description
    assert result is None


def test_svg_description_2():
    frame = Element.from_tag(
        '<draw:frame svg:width="10cm" svg:height="10cm" '
        'draw:z-index="0" draw:name="Another Frame" '
        'draw:style-name="Graphics" '
        'svg:x="10mm" svg:y="10mm" '
        'text:anchor-type="page" '
        'text:anchor-page-number="1"/>'
    )
    frame.svg_description = "some description"
    result = frame.svg_description
    assert result == "some description"


def test_svg_title_1():
    frame = Element.from_tag(
        '<draw:frame svg:width="10cm" svg:height="10cm" '
        'draw:z-index="0" draw:name="Another Frame" '
        'draw:style-name="Graphics" '
        'svg:x="10mm" svg:y="10mm" '
        'text:anchor-type="page" '
        'text:anchor-page-number="1"/>'
    )
    result = frame.svg_title
    assert result is None


def test_svg_title_2():
    frame = Element.from_tag(
        '<draw:frame svg:width="10cm" svg:height="10cm" '
        'draw:z-index="0" draw:name="Another Frame" '
        'draw:style-name="Graphics" '
        'svg:x="10mm" svg:y="10mm" '
        'text:anchor-type="page" '
        'text:anchor-page-number="1"/>'
    )
    frame.svg_title = "some title"
    result = frame.svg_title
    assert result == "some title"


def test_frame_get_formatted_text():
    frame = Frame.text_frame("content")
    expected = "  content\n  \n\n"
    assert frame.get_formatted_text() == expected


def test_frame_get_formatted_text_svg_title():
    frame = Frame()
    frame.svg_title = "some title"
    expected = "some title\n\n"
    assert frame.get_formatted_text() == expected


def test_frame_get_formatted_text_svg_description():
    frame = Frame()
    frame.svg_description = "some desc"
    expected = "some desc\n\n"
    assert frame.get_formatted_text() == expected


def test_frame_get_formatted_text_image():
    frame = Frame()
    frame.set_image("Pictures/image.jpg")
    expected = "[Image Pictures/image.jpg]\n\n"
    assert frame.get_formatted_text() == expected


def test_frame_get_formatted_text_image_rst():
    frame = Frame()
    frame.set_image("Pictures/image.jpg")
    expected = "\n.. image:: Pictures/image.jpg\n   :width: 37px\n   :height: 37px\n\n"
    assert frame.get_formatted_text({"rst_mode": True}) == expected


def test_frame_get_formatted_text_image_rst_no_size():
    frame = Frame()
    frame.set_image("Pictures/image.jpg")
    frame.width = None
    frame.height = None
    expected = "\n.. image:: Pictures/image.jpg\n\n"
    assert frame.get_formatted_text({"rst_mode": True}) == expected


def test_frame_get_formatted_text_image_rst_no_img():
    frame = Frame()
    frame.set_image("Pictures/image.jpg")
    expected = "|img1|\n"
    assert (
        frame.get_formatted_text(
            {"rst_mode": True, "no_img_level": True},
        )
        == expected
    )



#!/usr/bin/env python
"""Insert a circle and a lot of lines (a fractal) in a text document."""

from __future__ import annotations

import cmath
import os
from pathlib import Path

from odfdo import Document, EllipseShape, Header, LineShape, Paragraph

_DOC_SEQUENCE = 100
OUTPUT_DIR = Path(__file__).parent / "recipes_output" / "basic_drawing"
TARGET = "koch.odt"

CYCLES = 4  # beware, 5 is big, 6 is too big to display...


def save_new(document: Document, name: str) -> None:
    """Save a recipe result Document."""
    OUTPUT_DIR.mkdir(parents=True, exist_ok=True)
    new_path = OUTPUT_DIR / name
    print("Saving:", new_path)
    document.save(new_path, pretty=True)


def cm(x: float) -> str:
    """Return the value as cm string."""
    return f"{x:.2f}cm"


class Vector:
    """Vector class with Koch calculation."""

    def __init__(self, a: float | complex, b: float | complex) -> None:
        self.a = a
        self.b = b

    def koch_split(self) -> list[Vector]:
        c = self.a + 1.0 / 3.0 * (self.b - self.a)
        d = self.a + 2.0 / 3.0 * (self.b - self.a)
        m = 0.5 * (self.a + self.b)
        e = m + (d - c) * complex(0, -1)
        return [Vector(self.a, c), Vector(c, e), Vector(e, d), Vector(d, self.b)]

    def centimeter(self, index: int) -> tuple[str, str]:
        if index == 0:
            m = self.a
        else:
            m = self.b
        return (cm(m.real), cm(m.imag))


def koch(vector_list: list[Vector], cycles: int = 2) -> list[Vector]:
    """Generate a Koch fractal."""
    if cycles <= 0:
        return vector_list
    else:
        new_vector_list: list[Vector] = []
        for vector in vector_list:
            new_vector_list.extend(vector.koch_split())
        return koch(new_vector_list, cycles - 1)


def make_fractal_coords(
    side: float,
    vert_position: float,
) -> tuple[float | complex, list[Vector]]:
    """Return center and coordinates of a Koch fractal image."""
    orig = complex((17 - side) / 2.0, vert_position)
    v1 = Vector(orig, orig + complex(side, 0))
    v2 = Vector(v1.b, orig + cmath.rect(side, cmath.pi / 3))
    v3 = Vector(v2.b, orig)
    center: float | complex = (v1.a + v1.b + v2.b) / 3
    vector_list = koch([v1, v2, v3], cycles=CYCLES)
    return center, vector_list


def generate_document() -> Document:
    """Generate a document with image in it."""
    document = Document("text")
    body = document.body

    print("Making some Koch fractal")
    title = Header(1, "Some Koch fractal")
    body.append(title)

    style = document.get_style("graphic")
    style.set_properties({"svg:stroke_color": "#0000ff"})
    style.set_properties(fill_color="#ffffcc")

    paragraph = Paragraph("")
    body.append(paragraph)

    # some computation of oordinates
    center, vector_list = make_fractal_coords(side=12.0, vert_position=8.0)

    # create a circle
    radius = 8.0
    pos = center - complex(radius, radius)
    circle = EllipseShape(
        size=(cm(radius * 2), cm(radius * 2)),
        position=(cm(pos.real), cm(pos.imag)),
    )
    paragraph.append(circle)

    # create a drawing with a lot of lines
    paragraph.append(f"number of lines: {len(vector_list)}")
    for vector in vector_list:
        line = LineShape(p1=vector.centimeter(0), p2=vector.centimeter(1))
        paragraph.append(line)

    return document


def main() -> None:
    document = generate_document()
    test_unit(document)
    save_new(document, TARGET)


def test_unit(document: Document) -> None:
    # only for test suite:
    if "ODFDO_TESTING" not in os.environ:
        return

    assert len(document.body.paragraphs) == 2
    assert len(document.body.get_draw_lines()) == 768


if __name__ == "__main__":
    main()



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
"""Classes of Form controls like "form:text", "form:textarea", "form:password" ...

(The main objective of the current minimal implementation of forms is to parse
the existing form contents in a document.)"""

from __future__ import annotations

from decimal import Decimal
from typing import Any, ClassVar

from .element import Element, PropDef, PropDefBool, register_element_class
from .form_controls_mixins import (
    FormAsDictMixin,
    FormButtonTypeMixin,
    FormDelayRepeatMixin,
    FormImageAlignMixin,
    FormImagePositionMixin,
    FormMaxLengthMixin,
    FormSizetMixin,
    FormSourceListMixin,
    OfficeTargetFrameMixin,
)


class FormColumn(Element):
    """A column in a form grid control, "form:column".

    Attributes:
        name (str or None): The name of the column (form:name).
        control_implementation (str or None): The control implementation.
        label (str or None): The label of the column.
        text_style_name (str or None): The text style name (form:text-style-name).
    """

    _tag = "form:column"
    _properties: tuple[PropDef | PropDefBool, ...] = (
        PropDef("name", "form:name"),
        PropDef("control_implementation", "form:control-implementation"),
        PropDef("label", "form:label"),
        PropDef("text_style_name", "form:text-style-name"),
    )

    def __init__(
        self,
        name: str | None = None,
        control_implementation: str | None = None,
        label: str | None = None,
        text_style_name: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Create a FormColumn, "form:column".

        The "form:column" element is usable within the following element:
        "form:grid".

        Args:
            name: The name of the column.
            control_implementation: The control implementation.
            label: The label of the column.
            text_style_name: The text style name.
        """
        super().__init__(**kwargs)
        if self._do_init:
            if name is not None:
                self.name = name
            if control_implementation is not None:
                self.control_implementation = control_implementation
            if label is not None:
                self.label = label
            if text_style_name is not None:
                self.text_style_name = text_style_name

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} name={self.name}>"


FormColumn._define_attribut_property()


class FormGenericControl(Element):
    """An implementation-defined placeholder for a generic control, "form:generic-control".

    The generic control can contain any properties and any events.

    Attributes:
        name (str or None): The name of the control (form:name).
        control_implementation (str or None): The control implementation.
        xml_id (str or None): The unique XML ID (xml:id).
        xforms_bind (str or None): The XForms bind expression.
        form_id (str or None): The form ID (deprecated).
    """

    _tag = "form:generic-control"
    _properties: tuple[PropDef | PropDefBool, ...] = (
        PropDef("name", "form:name"),
        PropDef("control_implementation", "form:control-implementation"),
        PropDef("xml_id", "xml:id"),
        PropDef("xforms_bind", "xforms:bind"),
        PropDef("form_id", "form:id"),  # deprecated
    )

    def __init__(
        self,
        name: str | None = None,
        control_implementation: str | None = None,
        xml_id: str | None = None,
        xforms_bind: str | None = None,
        form_id: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Create a FormGenericControl, "form:generic-control".

        The "form:generic-control" element is usable within the following element:
        "form:form".

        Args:
            name: The name of the control.
            control_implementation: The control implementation.
            xml_id: The unique XML ID.
            xforms_bind: The XForms bind expression.
            form_id: The form ID (deprecated).
        """
        super().__init__(**kwargs)
        if self._do_init:
            if name is not None:
                self.name = name
            if control_implementation is not None:
                self.control_implementation = control_implementation
            if xml_id is not None:
                self.xml_id = xml_id
            if xforms_bind is not None:
                self.xforms_bind = xforms_bind
            if form_id is None:
                self.form_id = self.xml_id
            else:
                self.form_id = form_id

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} name={self.name} xml_id={self.xml_id}>"


FormGenericControl._define_attribut_property()


class FormHidden(FormAsDictMixin, FormGenericControl):
    """A control that does not have a visual representation, "form:hidden".

    Attributes:
        name (str or None): The name of the control (form:name).
        value (str or None): The value of the control.
        control_implementation (str or None): The control implementation.
        xml_id (str or None): The unique XML ID (xml:id).
        xforms_bind (str or None): The XForms bind expression.
        form_id (str or None): The form ID (deprecated).
    """

    _tag = "form:hidden"
    _properties: tuple[PropDef | PropDefBool, ...] = (
        PropDef("name", "form:name"),
        PropDef("value", "form:value"),
        PropDef("control_implementation", "form:control-implementation"),
        PropDef("xml_id", "xml:id"),
        PropDef("xforms_bind", "xforms:bind"),
        PropDef("form_id", "form:id"),  # deprecated
    )

    def __init__(
        self,
        name: str | None = None,
        value: str | None = None,
        control_implementation: str | None = None,
        xml_id: str | None = None,
        xforms_bind: str | None = None,
        form_id: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Create a FormHidden, "form:hidden".

        The "form:hidden" element is usable within the following element:
        "form:form".

        Args:
            name: The name of the control.
            value: The value of the control.
            control_implementation: The control implementation.
            xml_id: The unique XML ID.
            xforms_bind: The XForms bind expression.
            form_id: The form ID (deprecated).
        """
        super().__init__(
            name=name,
            control_implementation=control_implementation,
            xml_id=xml_id,
            xforms_bind=xforms_bind,
            form_id=form_id,
            **kwargs,
        )
        if self._do_init and value is not None:
            self.value = value

    def __str__(self) -> str:
        return ""


FormHidden._define_attribut_property()


class FormGrid(FormGenericControl):
    """A control a control that displays table data, "form:grid".

    Each column in the grid is specified by a "form:column" element.

    Attributes:
        name (str or None): The name of the control (form:name).
        control_implementation (str or None): The control implementation.
        title (str or None): The title or tooltip of the control.
        disabled (bool): Specifies if the control is disabled (form:disabled). Default is False.
        printable (bool): Specifies if the control is printable (form:printable). Default is True.
        tab_stop (bool): Specifies if the control is a tab stop (form:tab-stop). Default is True.
        tab_index (int or None): The tab order of the control (form:tab-index). Default is 0.
        xml_id (str or None): The unique XML ID (xml:id).
        xforms_bind (str or None): The XForms bind expression.
        form_id (str or None): The form ID (deprecated).
    """

    _tag = "form:grid"
    _properties: tuple[PropDef | PropDefBool, ...] = (
        PropDef("name", "form:name"),
        PropDef("control_implementation", "form:control-implementation"),
        PropDef("title", "form:title"),
        PropDefBool("disabled", "form:disabled", False),
        PropDefBool("printable", "form:printable", True),
        PropDefBool("tab_stop", "form:tab-stop", True),
        PropDef("xml_id", "xml:id"),
        PropDef("xforms_bind", "xforms:bind"),
        PropDef("form_id", "form:id"),  # deprecated
    )

    def __init__(
        self,
        name: str | None = None,
        control_implementation: str | None = None,
        title: str | None = None,
        disabled: bool | None = None,
        printable: bool | None = None,
        tab_index: int | None = None,
        tab_stop: bool | None = None,
        xml_id: str | None = None,
        xforms_bind: str | None = None,
        form_id: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Create a FormGrid, "form:grid".

        The "form:grid" element is usable within the following element:
        "form:form".

        Args:
            name: The name of the control.
            control_implementation: The control implementation.
            title: The title or tooltip.
            disabled: Specifies if the control is disabled.
            printable: Specifies if the control is printable.
            tab_index: The tab order of the control.
            tab_stop: Specifies if the control is a tab stop.
            xml_id: The unique XML ID.
            xforms_bind: The XForms bind expression.
            form_id: The form ID (deprecated).
        """
        super().__init__(
            name=name,
            control_implementation=control_implementation,
            xml_id=xml_id,
            xforms_bind=xforms_bind,
            form_id=form_id,
            **kwargs,
        )
        if self._do_init:
            if title is not None:
                self.title = title
            if disabled is not None:
                self.disabled = disabled
            if printable is not None:
                self.printable = printable
            self.tab_index = tab_index
            if tab_stop is not None:
                self.tab_stop = tab_stop

    @property
    def tab_index(self) -> int | None:
        return self._get_attribute_int_default("form:tab-index", 0)

    @tab_index.setter
    def tab_index(self, tab_index: int | None) -> None:
        self._set_attribute_int_default("form:tab-index", tab_index, 0)


FormGrid._define_attribut_property()


class FormText(FormAsDictMixin, FormMaxLengthMixin, FormGrid):
    """A control for displaying and inputting text on a single line, "form:text".

    Attributes:
        name (str or None): The name of the control (form:name).
        value (str or None): The value of the control.
        control_implementation (str or None): The control implementation.
        title (str or None): The title or tooltip of the control.
        disabled (bool): Specifies if the control is disabled (form:disabled). Default is False.
        printable (bool): Specifies if the control is printable (form:printable). Default is True.
        tab_stop (bool): Specifies if the control is a tab stop (form:tab-stop). Default is True.
        tab_index (int or None): The tab order of the control (form:tab-index). Default is 0.
        readonly (bool): Specifies if the control is read-only (form:readonly). Default is False.
        convert_empty_to_null (bool): Specifies if an empty value is converted to null (form:convert-empty-to-null). Default is False.
        current_value (str or None): The current value of the control (form:current-value).
        data_field (str or None): The data field used for storing the value.
        linked_cell (str or None): The linked spreadsheet cell.
        max_length (int or None): The maximum number of characters allowed (form:max-length).
        xml_id (str or None): The unique XML ID (xml:id).
        xforms_bind (str or None): The XForms bind expression.
        form_id (str or None): The form ID (deprecated).
    """

    _tag = "form:text"
    _properties: tuple[PropDef | PropDefBool, ...] = (
        PropDef("name", "form:name"),
        PropDef("value", "form:value"),
        PropDef("control_implementation", "form:control-implementation"),
        PropDef("title", "form:title"),
        PropDefBool("disabled", "form:disabled", False),
        PropDefBool("printable", "form:printable", True),
        PropDefBool("tab_stop", "form:tab-stop", True),
        PropDefBool("readonly", "form:readonly", False),
        PropDefBool("convert_empty_to_null", "form:convert-empty-to-null", False),
        PropDef("current_value", "form:current-value"),
        PropDef("data_field", "form:data-field"),
        PropDef("linked_cell", "form:linked-cell"),
        PropDef("xml_id", "xml:id"),
        PropDef("xforms_bind", "xforms:bind"),
        PropDef("form_id", "form:id"),  # deprecated
    )

    def __init__(
        self,
        name: str | None = None,
        value: str | None = None,
        control_implementation: str | None = None,
        title: str | None = None,
        disabled: bool | None = None,
        printable: bool | None = None,
        tab_index: int | None = None,
        tab_stop: bool | None = None,
        readonly: bool | None = None,
        convert_empty_to_null: bool | None = None,
        current_value: str | None = None,
        data_field: str | None = None,
        linked_cell: str | None = None,
        max_length: int | None = None,
        xml_id: str | None = None,
        xforms_bind: str | None = None,
        form_id: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Create a FormText, "form:text".

        The "form:text" element is usable within the following elements:
        "form:column" and "form:form".

        Args:
            name: The name of the control.
            value: The value of the control.
            control_implementation: The control implementation.
            title: The title or tooltip.
            disabled: Specifies if the control is disabled.
            printable: Specifies if the control is printable.
            tab_index: The tab order of the control.
            tab_stop: Specifies if the control is a tab stop.
            readonly: Specifies if the control is read-only.
            convert_empty_to_null: Specifies if an empty value is converted to null.
            current_value: The current value of the control.
            data_field: The data field used for storing the value.
            linked_cell: The linked spreadsheet cell.
            max_length: The maximum number of characters allowed.
            xml_id: The unique XML ID.
            xforms_bind: The XForms bind expression.
            form_id: The form ID (deprecated).
        """
        super().__init__(
            name=name,
            control_implementation=control_implementation,
            title=title,
            disabled=disabled,
            printable=printable,
            tab_index=tab_index,
            tab_stop=tab_stop,
            xml_id=xml_id,
            xforms_bind=xforms_bind,
            form_id=form_id,
            **kwargs,
        )
        if self._do_init:
            if value is not None:
                self.value = value
            if readonly is not None:
                self.readonly = readonly
            if convert_empty_to_null is not None:
                self.convert_empty_to_null = convert_empty_to_null
            if current_value is not None:
                self.current_value = current_value
            if data_field is not None:
                self.data_field = data_field
            if linked_cell is not None:
                self.linked_cell = linked_cell
            if max_length is not None:
                self.max_length = max_length

    def __str__(self) -> str:
        if self.current_value is not None:
            return self.current_value or ""
        return self.value or ""


FormText._define_attribut_property()


class FormTextarea(FormText):
    """A control for displaying and inputting text on multiple lines,
    "form:textarea".

    The "form:textarea" element may be used with plain-text values
    (specified by the form:current-value attribute) as well as with formatted
    text (specified as paragraph content). If both a form:current-value
    attribute and one or more "text:p" elements are present, it is
    implementation-dependent which text is used.

    The odfdo implementation returns str(self), fixme

    Attributes:
        name (str or None): The name of the control (form:name).
        value (str or None): The value of the control.
        control_implementation (str or None): The control implementation.
        title (str or None): The title or tooltip of the control.
        disabled (bool): Specifies if the control is disabled (form:disabled). Default is False.
        printable (bool): Specifies if the control is printable (form:printable). Default is True.
        tab_stop (bool): Specifies if the control is a tab stop (form:tab-stop). Default is True.
        tab_index (int or None): The tab order of the control (form:tab-index). Default is 0.
        readonly (bool): Specifies if the control is read-only (form:readonly). Default is False.
        convert_empty_to_null (bool): Specifies if an empty value is converted to null (form:convert-empty-to-null). Default is False.
        current_value (str or None): The current value of the control (form:current-value).
        data_field (str or None): The data field used for storing the value.
        linked_cell (str or None): The linked spreadsheet cell.
        max_length (int or None): The maximum number of characters allowed (form:max-length).
        xml_id (str or None): The unique XML ID (xml:id).
        xforms_bind (str or None): The XForms bind expression.
        form_id (str or None): The form ID (deprecated).
    """

    _tag = "form:textarea"
    _properties: tuple[PropDef | PropDefBool, ...] = (
        PropDef("name", "form:name"),
        PropDef("value", "form:value"),
        PropDef("control_implementation", "form:control-implementation"),
        PropDef("title", "form:title"),
        PropDefBool("disabled", "form:disabled", False),
        PropDefBool("printable", "form:printable", True),
        PropDefBool("tab_stop", "form:tab-stop", True),
        PropDefBool("readonly", "form:readonly", False),
        PropDefBool("convert_empty_to_null", "form:convert-empty-to-null", False),
        PropDef("current_value", "form:current-value"),
        PropDef("data_field", "form:data-field"),
        PropDef("linked_cell", "form:linked-cell"),
        PropDef("xml_id", "xml:id"),
        PropDef("xforms_bind", "xforms:bind"),
        PropDef("form_id", "form:id"),  # deprecated#
    )

    def __init__(
        self,
        name: str | None = None,
        value: str | None = None,
        control_implementation: str | None = None,
        title: str | None = None,
        disabled: bool | None = None,
        printable: bool | None = None,
        tab_index: int | None = None,
        tab_stop: bool | None = None,
        readonly: bool | None = None,
        convert_empty_to_null: bool | None = None,
        current_value: str | None = None,
        data_field: str | None = None,
        linked_cell: str | None = None,
        max_length: int | None = None,
        xml_id: str | None = None,
        xforms_bind: str | None = None,
        form_id: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Create a FormTextarea, "form:textarea".

        The "form:textarea" element is usable within the following elements:
        "form:column" and "form:form".

        Args:
            name: The name of the control.
            value: The value of the control.
            control_implementation: The control implementation.
            title: The title or tooltip.
            disabled: Specifies if the control is disabled.
            printable: Specifies if the control is printable.
            tab_index: The tab order of the control.
            tab_stop: Specifies if the control is a tab stop.
            readonly: Specifies if the control is read-only.
            convert_empty_to_null: Specifies if an empty value is converted to null.
            current_value: The current value of the control.
            data_field: The data field used for storing the value.
            linked_cell: The linked spreadsheet cell.
            max_length: The maximum number of characters allowed.
            xml_id: The unique XML ID.
            xforms_bind: The XForms bind expression.
            form_id: The form ID (deprecated).
        """
        super().__init__(
            name=name,
            value=value,
            control_implementation=control_implementation,
            title=title,
            disabled=disabled,
            printable=printable,
            tab_index=tab_index,
            tab_stop=tab_stop,
            readonly=readonly,
            convert_empty_to_null=convert_empty_to_null,
            current_value=current_value,
            data_field=data_field,
            linked_cell=linked_cell,
            max_length=max_length,
            xml_id=xml_id,
            xforms_bind=xforms_bind,
            form_id=form_id,
            **kwargs,
        )

    def __str__(self) -> str:
        return self.inner_text


FormTextarea._define_attribut_property()


class FormPassword(FormAsDictMixin, FormMaxLengthMixin, FormGrid):
    """A control that uses an echo character to hide password input by a user, "form:password".

    Attributes:
        name (str or None): The name of the control (form:name).
        value (str or None): The value of the control.
        control_implementation (str or None): The control implementation.
        title (str or None): The title or tooltip of the control.
        echo_char (str): The character used to echo password input (form:echo-char). Default is "*".
        disabled (bool): Specifies if the control is disabled (form:disabled). Default is False.
        printable (bool): Specifies if the control is printable (form:printable). Default is True.
        tab_stop (bool): Specifies if the control is a tab stop (form:tab-stop). Default is True.
        tab_index (int or None): The tab order of the control (form:tab-index). Default is 0.
        convert_empty_to_null (bool): Specifies if an empty value is converted to null (form:convert-empty-to-null). Default is False.
        linked_cell (str or None): The linked spreadsheet cell.
        max_length (int or None): The maximum number of characters allowed (form:max-length).
        xml_id (str or None): The unique XML ID (xml:id).
        xforms_bind (str or None): The XForms bind expression.
        form_id (str or None): The form ID (deprecated).
    """

    _tag = "form:password"
    _properties: tuple[PropDef | PropDefBool, ...] = (
        PropDef("name", "form:name"),
        PropDef("value", "form:value"),
        PropDef("control_implementation", "form:control-implementation"),
        PropDef("title", "form:title"),
        PropDef("echo_char", "form:echo-char"),
        PropDefBool("disabled", "form:disabled", False),
        PropDefBool("printable", "form:printable", True),
        PropDefBool("tab_stop", "form:tab-stop", True),
        PropDefBool("convert_empty_to_null", "form:convert-empty-to-null", False),
        PropDef("linked_cell", "form:linked-cell"),
        PropDef("xml_id", "xml:id"),
        PropDef("xforms_bind", "xforms:bind"),
        PropDef("form_id", "form:id"),  # deprecated
    )

    def __init__(
        self,
        name: str | None = None,
        value: str | None = None,
        control_implementation: str | None = None,
        title: str | None = None,
        echo_char: str | None = "*",
        disabled: bool | None = None,
        printable: bool | None = None,
        tab_index: int | None = None,
        tab_stop: bool | None = None,
        convert_empty_to_null: bool | None = None,
        linked_cell: str | None = None,
        max_length: int | None = None,
        xml_id: str | None = None,
        xforms_bind: str | None = None,
        form_id: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Create a FormPassword, "form:password".

        The "form:password" element is usable within the following element:
        "form:form".

        Args:
            name: The name of the control.
            value: The value of the control.
            control_implementation: The control implementation.
            title: The title or tooltip.
            echo_char: The character used to echo password input. Defaults to "*".
            disabled: Specifies if the control is disabled.
            printable: Specifies if the control is printable.
            tab_index: The tab order of the control.
            tab_stop: Specifies if the control is a tab stop.
            convert_empty_to_null: Specifies if an empty value is converted to null.
            linked_cell: The linked spreadsheet cell.
            max_length: The maximum number of characters allowed.
            xml_id: The unique XML ID.
            xforms_bind: The XForms bind expression.
            form_id: The form ID (deprecated).
        """
        super().__init__(
            name=name,
            control_implementation=control_implementation,
            title=title,
            disabled=disabled,
            printable=printable,
            tab_index=tab_index,
            tab_stop=tab_stop,
            max_length=max_length,
            xml_id=xml_id,
            xforms_bind=xforms_bind,
            form_id=form_id,
            **kwargs,
        )
        if self._do_init:
            if value is not None:
                self.value = value
            self.echo_char = echo_char or "*"
            if convert_empty_to_null is not None:
                self.convert_empty_to_null = convert_empty_to_null
            if linked_cell is not None:
                self.linked_cell = linked_cell
            if max_length is not None:
                self.max_length = max_length

    def __str__(self) -> str:
        return self.echo_char or ""


FormPassword._define_attribut_property()


class FormFile(FormAsDictMixin, FormMaxLengthMixin, FormGrid):
    """A control for or selecting a file, "form:file".

    Attributes:
        name (str or None): The name of the control (form:name).
        value (str or None): The value of the control.
        control_implementation (str or None): The control implementation.
        title (str or None): The title or tooltip of the control.
        disabled (bool): Specifies if the control is disabled (form:disabled). Default is False.
        printable (bool): Specifies if the control is printable (form:printable). Default is True.
        tab_stop (bool): Specifies if the control is a tab stop (form:tab-stop). Default is True.
        tab_index (int or None): The tab order of the control (form:tab-index). Default is 0.
        readonly (bool): Specifies if the control is read-only (form:readonly). Default is False.
        current_value (str or None): The current value of the control (form:current-value).
        linked_cell (str or None): The linked spreadsheet cell.
        max_length (int or None): The maximum number of characters allowed (form:max-length).
        xml_id (str or None): The unique XML ID (xml:id).
        xforms_bind (str or None): The XForms bind expression.
        form_id (str or None): The form ID (deprecated).
    """

    _tag = "form:file"
    _properties: tuple[PropDef | PropDefBool, ...] = (
        PropDef("name", "form:name"),
        PropDef("value", "form:value"),
        PropDef("control_implementation", "form:control-implementation"),
        PropDef("title", "form:title"),
        PropDefBool("disabled", "form:disabled", False),
        PropDefBool("printable", "form:printable", True),
        PropDefBool("tab_stop", "form:tab-stop", True),
        PropDefBool("readonly", "form:readonly", False),
        PropDef("current_value", "form:current-value"),
        PropDef("linked_cell", "form:linked-cell"),
        PropDef("xml_id", "xml:id"),
        PropDef("xforms_bind", "xforms:bind"),
        PropDef("form_id", "form:id"),  # deprecated
    )

    def __init__(
        self,
        name: str | None = None,
        value: str | None = None,
        control_implementation: str | None = None,
        title: str | None = None,
        disabled: bool | None = None,
        printable: bool | None = None,
        tab_index: int | None = None,
        tab_stop: bool | None = None,
        readonly: bool | None = None,
        current_value: str | None = None,
        linked_cell: str | None = None,
        max_length: int | None = None,
        xml_id: str | None = None,
        xforms_bind: str | None = None,
        form_id: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Create a FormFile, "form:file".

        The "form:file" element is usable within the following element:
        "form:form".

        Args:
            name: The name of the control.
            value: The value of the control.
            control_implementation: The control implementation.
            title: The title or tooltip.
            disabled: Specifies if the control is disabled.
            printable: Specifies if the control is printable.
            tab_index: The tab order of the control.
            tab_stop: Specifies if the control is a tab stop.
            readonly: Specifies if the control is read-only.
            current_value: The current value of the control.
            linked_cell: The linked spreadsheet cell.
            max_length: The maximum number of characters allowed.
            xml_id: The unique XML ID.
            xforms_bind: The XForms bind expression.
            form_id: The form ID (deprecated).
        """
        super().__init__(
            name=name,
            control_implementation=control_implementation,
            title=title,
            disabled=disabled,
            printable=printable,
            tab_index=tab_index,
            tab_stop=tab_stop,
            max_length=max_length,
            xml_id=xml_id,
            xforms_bind=xforms_bind,
            form_id=form_id,
            **kwargs,
        )
        if self._do_init:
            if value is not None:
                self.value = value
            if readonly is not None:
                self.readonly = readonly
            if current_value is not None:
                self.current_value = current_value
            if linked_cell is not None:
                self.linked_cell = linked_cell
            if max_length is not None:
                self.max_length = max_length

    def __str__(self) -> str:
        if self.current_value is not None:
            return self.current_value or ""
        return self.value or ""


FormFile._define_attribut_property()


class FormFormattedText(FormDelayRepeatMixin, FormText):
    """A control for inputting text, which follows the format defined by a
    data style that is assigned to the control's graphical shape,
    "form:formatted-text"

    Attributes:
        name (str or None): The name of the control (form:name).
        value (str or None): The value of the control.
        control_implementation (str or None): The control implementation.
        title (str or None): The title or tooltip of the control.
        disabled (bool): Specifies if the control is disabled (form:disabled). Default is False.
        printable (bool): Specifies if the control is printable (form:printable). Default is True.
        tab_stop (bool): Specifies if the control is a tab stop (form:tab-stop). Default is True.
        tab_index (int or None): The tab order of the control (form:tab-index). Default is 0.
        readonly (bool): Specifies if the control is read-only (form:readonly). Default is False.
        convert_empty_to_null (bool): Specifies if an empty value is converted to null (form:convert-empty-to-null). Default is False.
        current_value (str or None): The current value of the control (form:current-value).
        data_field (str or None): The data field used for storing the value.
        repeat (bool or None): Specifies if the control repeats when a button is clicked (form:repeat).
        delay_for_repeat (str or None): The delay for repeating the action (form:delay-for-repeat).
        linked_cell (str or None): The linked spreadsheet cell.
        max_length (int or None): The maximum number of characters allowed (form:max-length).
        min_value (str or None): The minimum numeric value allowed (form:min-value).
        max_value (str or None): The maximum numeric value allowed (form:max-value).
        spin_button (bool): Specifies if a spin button is available (form:spin-button). Default is False.
        validation (bool): Specifies if the control performs validation (form:validation). Default is False.
        xml_id (str or None): The unique XML ID (xml:id).
        xforms_bind (str or None): The XForms bind expression.
        form_id (str or None): The form ID (deprecated).
    """

    _tag = "form:formatted-text"
    _properties: tuple[PropDef | PropDefBool, ...] = (
        PropDef("name", "form:name"),
        PropDef("value", "form:value"),
        PropDef("control_implementation", "form:control-implementation"),
        PropDef("title", "form:title"),
        PropDefBool("disabled", "form:disabled", False),
        PropDefBool("printable", "form:printable", True),
        PropDefBool("tab_stop", "form:tab-stop", True),
        PropDefBool("readonly", "form:readonly", False),
        PropDefBool("convert_empty_to_null", "form:convert-empty-to-null", False),
        PropDef("current_value", "form:current-value"),
        PropDef("data_field", "form:data-field"),
        PropDef("repeat", "form:repeat"),
        PropDef("linked_cell", "form:linked-cell"),
        PropDef("min_value", "form:min-value"),
        PropDef("max_value", "form:max-value"),
        PropDefBool("spin_button", "form:spin-button", False),
        PropDefBool("validation", "form:validation", False),
        PropDef("xml_id", "xml:id"),
        PropDef("xforms_bind", "xforms:bind"),
        PropDef("form_id", "form:id"),  # deprecated
    )

    def __init__(
        self,
        name: str | None = None,
        value: str | None = None,
        control_implementation: str | None = None,
        title: str | None = None,
        disabled: bool | None = None,
        printable: bool | None = None,
        tab_index: int | None = None,
        tab_stop: bool | None = None,
        readonly: bool | None = None,
        convert_empty_to_null: bool | None = None,
        current_value: str | None = None,
        data_field: str | None = None,
        repeat: bool | None = None,
        delay_for_repeat: str | None = None,
        linked_cell: str | None = None,
        max_length: int | None = None,
        min_value: str | None = None,
        max_value: str | None = None,
        spin_button: bool | None = None,
        validation: bool | None = None,
        xml_id: str | None = None,
        xforms_bind: str | None = None,
        form_id: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Create a FormFormattedText, "form:formatted-text".

        The "form:formatted-text" element is usable within the following elements:
        "form:column" and "form:form".

        Args:
            name: The name of the control.
            value: The value of the control.
            control_implementation: The control implementation.
            title: The title or tooltip.
            disabled: Specifies if the control is disabled.
            printable: Specifies if the control is printable.
            tab_index: The tab order of the control.
            tab_stop: Specifies if the control is a tab stop.
            readonly: Specifies if the control is read-only.
            convert_empty_to_null: Specifies if an empty value is converted to null.
            current_value: The current value of the control.
            data_field: The data field used for storing the value.
            repeat: Specifies if the control repeats when a button is clicked.
            delay_for_repeat: The delay for repeating the action.
            linked_cell: The linked spreadsheet cell.
            max_length: The maximum number of characters allowed.
            min_value: The minimum numeric value allowed.
            max_value: The maximum numeric value allowed.
            spin_button: Specifies if a spin button is available.
            validation: Specifies if the control performs validation.
            xml_id: The unique XML ID.
            xforms_bind: The XForms bind expression.
            form_id: The form ID (deprecated).
        """
        super().__init__(
            name=name,
            control_implementation=control_implementation,
            title=title,
            value=value,
            disabled=disabled,
            printable=printable,
            readonly=readonly,
            convert_empty_to_null=convert_empty_to_null,
            current_value=current_value,
            data_field=data_field,
            linked_cell=linked_cell,
            tab_index=tab_index,
            tab_stop=tab_stop,
            max_length=max_length,
            xml_id=xml_id,
            xforms_bind=xforms_bind,
            form_id=form_id,
            **kwargs,
        )
        if self._do_init:
            if repeat is not None:
                self.repeat = repeat
            if delay_for_repeat is not None:
                self.delay_for_repeat = delay_for_repeat
            if min_value is not None:
                self.min_value = min_value
            if max_value is not None:
                self.max_value = max_value
            if spin_button is not None:
                self.spin_button = spin_button
            if validation is not None:
                self.validation = validation


FormFormattedText._define_attribut_property()


class FormNumber(FormDelayRepeatMixin, FormAsDictMixin, FormMaxLengthMixin, FormGrid):
    """A control which allows the user to enter a floating-point number,
    "form:number".

    Attributes:
        name (str or None): The name of the control (form:name).
        value (str or None): The value of the control (form:value).
        control_implementation (str or None): The control implementation (form:control-implementation).
        title (str or None): The title or tooltip of the control (form:title).
        disabled (bool): Specifies if the control is disabled (form:disabled).
        printable (bool): Specifies if the control is printable (form:printable).
        tab_stop (bool): Specifies if the control is a tab stop (form:tab-stop).
        readonly (bool): Specifies if the control is read-only (form:readonly).
        convert_empty_to_null (bool): Specifies if an empty value is converted to null (form:convert-empty-to-null).
        data_field (str or None): The data field used for storing the value (form:data-field).
        repeat (bool or None): Specifies if the control repeats when a button is clicked (form:repeat).
        linked_cell (str or None): The linked spreadsheet cell (form:linked-cell).
        spin_button (bool): Specifies if a spin button is available (form:spin-button).
        xml_id (str or None): The unique XML ID (xml:id).
        xforms_bind (str or None): The XForms bind expression (xforms:bind).
        form_id (str or None): The form ID (deprecated, form:id).
        current_value (Decimal or int or None): The current value of the control (form:current-value).
        min_value (Decimal or int or None): The minimum numeric value allowed (form:min-value).
        max_value (Decimal or int or None): The maximum numeric value allowed (form:max-value).
        tab_index (int or None): The tab order of the control (form:tab-index).
        max_length (int or None): The maximum number of characters allowed (form:max-length).
        delay_for_repeat (str or None): The delay for repeating the action (form:delay-for-repeat).
    """

    _tag = "form:number"
    _properties: tuple[PropDef | PropDefBool, ...] = (
        PropDef("name", "form:name"),
        PropDef("value", "form:value"),
        PropDef("control_implementation", "form:control-implementation"),
        PropDef("title", "form:title"),
        PropDefBool("disabled", "form:disabled", False),
        PropDefBool("printable", "form:printable", True),
        PropDefBool("tab_stop", "form:tab-stop", True),
        PropDefBool("readonly", "form:readonly", False),
        PropDefBool("convert_empty_to_null", "form:convert-empty-to-null", False),
        PropDef("data_field", "form:data-field"),
        PropDef("repeat", "form:repeat"),
        PropDef("linked_cell", "form:linked-cell"),
        PropDefBool("spin_button", "form:spin-button", False),
        PropDef("xml_id", "xml:id"),
        PropDef("xforms_bind", "xforms:bind"),
        PropDef("form_id", "form:id"),  # deprecated
    )

    def __init__(
        self,
        name: str | None = None,
        value: str | None = None,
        control_implementation: str | None = None,
        title: str | None = None,
        disabled: bool | None = None,
        printable: bool | None = None,
        tab_index: int | None = None,
        tab_stop: bool | None = None,
        readonly: bool | None = None,
        convert_empty_to_null: bool | None = None,
        current_value: Decimal | int | float | None = None,
        data_field: str | None = None,
        repeat: bool | None = None,
        delay_for_repeat: str | None = None,
        linked_cell: str | None = None,
        max_length: int | None = None,
        min_value: Decimal | int | float | None = None,
        max_value: Decimal | int | float | None = None,
        spin_button: bool | None = None,
        xml_id: str | None = None,
        xforms_bind: str | None = None,
        form_id: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Create a FormNumber, "form:number".

        The "form:number" element is usable within the following elements:
        "form:column" and "form:form".

        Args:
            name: The name of the control.
            value: The value of the control.
            control_implementation: The control implementation.
            title: The title or tooltip of the control.
            disabled: If True, the control is disabled.
            printable: If True, the control is printable.
            tab_index: The tab order of the control.
            tab_stop: If True, the control is a tab stop.
            readonly: If True, the control is read-only.
            convert_empty_to_null: If True, an empty value is converted to null.
            current_value: The current value of the control.
            data_field: The data field for storing the value.
            repeat: If True, the control repeats when a button is clicked.
            delay_for_repeat: The delay for repeating the action. Defaults to "PT0.050S".
            linked_cell: The linked spreadsheet cell.
            max_length: The maximum number of characters allowed.
            min_value: The minimum value allowed.
            max_value: The maximum value allowed.
            spin_button: If True, a spin button is available.
            xml_id: The unique XML ID.
            xforms_bind: The XForms bind expression.
            form_id: The form ID (deprecated).
        """
        super().__init__(
            name=name,
            control_implementation=control_implementation,
            title=title,
            disabled=disabled,
            printable=printable,
            tab_index=tab_index,
            tab_stop=tab_stop,
            max_length=max_length,
            xml_id=xml_id,
            xforms_bind=xforms_bind,
            form_id=form_id,
            **kwargs,
        )
        if self._do_init:
            if value is not None:
                self.value = value
            if readonly is not None:
                self.readonly = readonly
            if convert_empty_to_null is not None:
                self.convert_empty_to_null = convert_empty_to_null
            if current_value is not None:
                self.current_value = current_value
            if data_field is not None:
                self.data_field = data_field
            if linked_cell is not None:
                self.linked_cell = linked_cell
            if repeat is not None:
                self.repeat = repeat
            if delay_for_repeat is not None:
                self.delay_for_repeat = delay_for_repeat
            if min_value is not None:
                self.min_value = min_value
            if max_value is not None:
                self.max_value = max_value
            if spin_button is not None:
                self.spin_button = spin_button

    @property
    def current_value(self) -> Decimal | int | None:
        return self.get_attribute_number("form:current-value")

    @current_value.setter
    def current_value(self, current_value: Decimal | int | float | None) -> None:
        self._set_attribute_number_default("form:current-value", current_value, None)

    @property
    def min_value(self) -> Decimal | int | None:
        return self.get_attribute_number("form:min-value")

    @min_value.setter
    def min_value(self, min_value: Decimal | int | float | None) -> None:
        self._set_attribute_number_default("form:min-value", min_value, None)

    @property
    def max_value(self) -> Decimal | int | None:
        return self.get_attribute_number("form:max-value")

    @max_value.setter
    def max_value(self, max_value: Decimal | int | float | None) -> None:
        self._set_attribute_number_default("form:max-value", max_value, None)

    def __str__(self) -> str:
        if self.current_value is not None:
            return str(self.current_value)
        return self.value or ""


FormNumber._define_attribut_property()


class FormDate(FormDelayRepeatMixin, FormText):
    """A control for inputting date data, "form:date".

    Attributes:
        name (str or None): The name of the control (form:name).
        value (str or None): The value of the control (form:value).
        control_implementation (str or None): The control implementation (form:control-implementation).
        title (str or None): The title or tooltip of the control (form:title).
        disabled (bool): Specifies if the control is disabled (form:disabled).
        printable (bool): Specifies if the control is printable (form:printable).
        tab_stop (bool): Specifies if the control is a tab stop (form:tab-stop).
        readonly (bool): Specifies if the control is read-only (form:readonly).
        convert_empty_to_null (bool): Specifies if an empty value is converted to null (form:convert-empty-to-null).
        current_value (str or None): The current value of the control (form:current-value).
        data_field (str or None): The data field used for storing the value (form:data-field).
        repeat (bool or None): Specifies if the control repeats when a button is clicked (form:repeat).
        linked_cell (str or None): The linked spreadsheet cell (form:linked-cell).
        min_value (str or None): The minimum date value allowed (form:min-value).
        max_value (str or None): The maximum date value allowed (form:max-value).
        spin_button (bool): Specifies if a spin button is available (form:spin-button).
        xml_id (str or None): The unique XML ID (xml:id).
        xforms_bind (str or None): The XForms bind expression (xforms:bind).
        form_id (str or None): The form ID (deprecated, form:id).
        tab_index (int or None): The tab order of the control (form:tab-index).
        max_length (int or None): The maximum number of characters allowed (form:max-length).
        delay_for_repeat (str or None): The delay for repeating the action (form:delay-for-repeat).
    """

    _tag = "form:date"
    _properties: tuple[PropDef | PropDefBool, ...] = (
        PropDef("name", "form:name"),
        PropDef("value", "form:value"),
        PropDef("control_implementation", "form:control-implementation"),
        PropDef("title", "form:title"),
        PropDefBool("disabled", "form:disabled", False),
        PropDefBool("printable", "form:printable", True),
        PropDefBool("tab_stop", "form:tab-stop", True),
        PropDefBool("readonly", "form:readonly", False),
        PropDefBool("convert_empty_to_null", "form:convert-empty-to-null", False),
        PropDef("current_value", "form:current-value"),
        PropDef("data_field", "form:data-field"),
        PropDef("repeat", "form:repeat"),
        PropDef("linked_cell", "form:linked-cell"),
        PropDef("min_value", "form:min-value"),
        PropDef("max_value", "form:max-value"),
        PropDefBool("spin_button", "form:spin-button", False),
        PropDef("xml_id", "xml:id"),
        PropDef("xforms_bind", "xforms:bind"),
        PropDef("form_id", "form:id"),  # deprecated
    )

    def __init__(
        self,
        name: str | None = None,
        value: str | None = None,
        control_implementation: str | None = None,
        title: str | None = None,
        disabled: bool | None = None,
        printable: bool | None = None,
        tab_index: int | None = None,
        tab_stop: bool | None = None,
        readonly: bool | None = None,
        convert_empty_to_null: bool | None = None,
        current_value: str | None = None,
        data_field: str | None = None,
        repeat: bool | None = None,
        delay_for_repeat: str | None = None,
        linked_cell: str | None = None,
        max_length: int | None = None,
        min_value: str | None = None,
        max_value: str | None = None,
        spin_button: bool | None = None,
        xml_id: str | None = None,
        xforms_bind: str | None = None,
        form_id: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Create a FormDate, "form:date".

        The "form:date" element is usable within the following elements:
        "form:column" and "form:form".

        Args:
            name: The name of the control.
            value: The value of the control.
            control_implementation: The control implementation.
            title: The title or tooltip of the control.
            disabled: If True, the control is disabled.
            printable: If True, the control is printable.
            tab_index: The tab order of the control.
            tab_stop: If True, the control is a tab stop.
            readonly: If True, the control is read-only.
            convert_empty_to_null: If True, an empty value is converted to null.
            current_value: The current value of the control.
            data_field: The data field for storing the value.
            repeat: If True, the control repeats when a button is clicked.
            delay_for_repeat: The delay for repeating the action. Defaults to "PT0.050S".
            linked_cell: The linked spreadsheet cell.
            max_length: The maximum number of characters allowed.
            min_value: The minimum value allowed.
            max_value: The maximum value allowed.
            spin_button: If True, a spin button is available.
            xml_id: The unique XML ID.
            xforms_bind: The XForms bind expression.
            form_id: The form ID (deprecated).
        """
        super().__init__(
            name=name,
            control_implementation=control_implementation,
            title=title,
            value=value,
            disabled=disabled,
            printable=printable,
            readonly=readonly,
            convert_empty_to_null=convert_empty_to_null,
            current_value=current_value,
            data_field=data_field,
            linked_cell=linked_cell,
            tab_index=tab_index,
            tab_stop=tab_stop,
            max_length=max_length,
            xml_id=xml_id,
            xforms_bind=xforms_bind,
            form_id=form_id,
            **kwargs,
        )
        if self._do_init:
            if repeat is not None:
                self.repeat = repeat
            if delay_for_repeat is not None:
                self.delay_for_repeat = delay_for_repeat
            if min_value is not None:
                self.min_value = min_value
            if max_value is not None:
                self.max_value = max_value
            if spin_button is not None:
                self.spin_button = spin_button


FormDate._define_attribut_property()


class FormTime(FormDelayRepeatMixin, FormText):
    """A control for inputting time data, "form:time".

    Attributes:
        name (str or None): The name of the control (form:name).
        value (str or None): The value of the control (form:value).
        control_implementation (str or None): The control implementation (form:control-implementation).
        title (str or None): The title or tooltip of the control (form:title).
        disabled (bool): Specifies if the control is disabled (form:disabled).
        printable (bool): Specifies if the control is printable (form:printable).
        tab_stop (bool): Specifies if the control is a tab stop (form:tab-stop).
        readonly (bool): Specifies if the control is read-only (form:readonly).
        convert_empty_to_null (bool): Specifies if an empty value is converted to null (form:convert-empty-to-null).
        current_value (str or None): The current value of the control (form:current-value).
        data_field (str or None): The data field used for storing the value (form:data-field).
        repeat (bool or None): Specifies if the control repeats when a button is clicked (form:repeat).
        linked_cell (str or None): The linked spreadsheet cell (form:linked-cell).
        min_value (str or None): The minimum time value allowed (form:min-value).
        max_value (str or None): The maximum time value allowed (form:max-value).
        spin_button (bool): Specifies if a spin button is available (form:spin-button).
        xml_id (str or None): The unique XML ID (xml:id).
        xforms_bind (str or None): The XForms bind expression (xforms:bind).
        form_id (str or None): The form ID (deprecated, form:id).
        tab_index (int or None): The tab order of the control (form:tab-index).
        max_length (int or None): The maximum number of characters allowed (form:max-length).
        delay_for_repeat (str or None): The delay for repeating the action (form:delay-for-repeat).
    """

    _tag = "form:time"
    _properties: tuple[PropDef | PropDefBool, ...] = (
        PropDef("name", "form:name"),
        PropDef("value", "form:value"),
        PropDef("control_implementation", "form:control-implementation"),
        PropDef("title", "form:title"),
        PropDefBool("disabled", "form:disabled", False),
        PropDefBool("printable", "form:printable", True),
        PropDefBool("tab_stop", "form:tab-stop", True),
        PropDefBool("readonly", "form:readonly", False),
        PropDefBool("convert_empty_to_null", "form:convert-empty-to-null", False),
        PropDef("current_value", "form:current-value"),
        PropDef("data_field", "form:data-field"),
        PropDef("repeat", "form:repeat"),
        PropDef("linked_cell", "form:linked-cell"),
        PropDef("min_value", "form:min-value"),
        PropDef("max_value", "form:max-value"),
        PropDefBool("spin_button", "form:spin-button", False),
        PropDef("xml_id", "xml:id"),
        PropDef("xforms_bind", "xforms:bind"),
        PropDef("form_id", "form:id"),  # deprecated
    )

    def __init__(
        self,
        name: str | None = None,
        value: str | None = None,
        control_implementation: str | None = None,
        title: str | None = None,
        disabled: bool | None = None,
        printable: bool | None = None,
        tab_index: int | None = None,
        tab_stop: bool | None = None,
        readonly: bool | None = None,
        convert_empty_to_null: bool | None = None,
        current_value: str | None = None,
        data_field: str | None = None,
        repeat: bool | None = None,
        delay_for_repeat: str | None = None,
        linked_cell: str | None = None,
        max_length: int | None = None,
        min_value: str | None = None,
        max_value: str | None = None,
        spin_button: bool | None = None,
        xml_id: str | None = None,
        xforms_bind: str | None = None,
        form_id: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Create a FormTime, "form:time".

        The "form:time" element is usable within the following elements:
        "form:column" and "form:form".

        Args:
            name: The name of the control.
            value: The value of the control.
            control_implementation: The control implementation.
            title: The title or tooltip of the control.
            disabled: If True, the control is disabled.
            printable: If True, the control is printable.
            tab_index: The tab order of the control.
            tab_stop: If True, the control is a tab stop.
            readonly: If True, the control is read-only.
            convert_empty_to_null: If True, an empty value is converted to null.
            current_value: The current value of the control.
            data_field: The data field for storing the value.
            repeat: If True, the control repeats when a button is clicked.
            delay_for_repeat: The delay for repeating the action. Defaults to "PT0.050S".
            linked_cell: The linked spreadsheet cell.
            max_length: The maximum number of characters allowed.
            min_value: The minimum value allowed.
            max_value: The maximum value allowed.
            spin_button: If True, a spin button is available.
            xml_id: The unique XML ID.
            xforms_bind: The XForms bind expression.
            form_id: The form ID (deprecated).
        """
        super().__init__(
            name=name,
            control_implementation=control_implementation,
            title=title,
            value=value,
            disabled=disabled,
            printable=printable,
            readonly=readonly,
            convert_empty_to_null=convert_empty_to_null,
            current_value=current_value,
            data_field=data_field,
            linked_cell=linked_cell,
            tab_index=tab_index,
            tab_stop=tab_stop,
            max_length=max_length,
            xml_id=xml_id,
            xforms_bind=xforms_bind,
            form_id=form_id,
            **kwargs,
        )
        if self._do_init:
            if repeat is not None:
                self.repeat = repeat
            if delay_for_repeat is not None:
                self.delay_for_repeat = delay_for_repeat
            if min_value is not None:
                self.min_value = min_value
            if max_value is not None:
                self.max_value = max_value
            if spin_button is not None:
                self.spin_button = spin_button


FormTime._define_attribut_property()


class FormFixedText(FormGenericControl):
    """A control which attaches additional information to controls, or
    displays information, "form:fixed-text"

    Only one label may be associated with a control.

    Attributes:
        name (str or None): The name of the control (form:name).
        label (str or None): The label of the control (form:label).
        title (str or None): The title or tooltip of the control (form:title).
        form_for (str or None): The ID of the control this label is for (form:for).
        multi_line (bool): Specifies if the label can have multiple lines (form:multi-line).
        control_implementation (str or None): The control implementation (form:control-implementation).
        disabled (bool): Specifies if the control is disabled (form:disabled).
        printable (bool): Specifies if the control is printable (form:printable).
        xml_id (str or None): The unique XML ID (xml:id).
        xforms_bind (str or None): The XForms bind expression (xforms:bind).
        form_id (str or None): The form ID (deprecated, form:id).
    """

    _tag = "form:fixed-text"
    _properties: tuple[PropDef | PropDefBool, ...] = (
        PropDef("name", "form:name"),
        PropDef("label", "form:label"),
        PropDef("title", "form:title"),
        PropDef("form_for", "form:for"),
        PropDefBool("multi_line", "form:multi-line", False),
        PropDef("control_implementation", "form:control-implementation"),
        PropDefBool("disabled", "form:disabled", False),
        PropDefBool("printable", "form:printable", True),
        PropDef("xml_id", "xml:id"),
        PropDef("xforms_bind", "xforms:bind"),
        PropDef("form_id", "form:id"),  # deprecated
    )

    def __init__(
        self,
        name: str | None = None,
        label: str | None = None,
        title: str | None = None,
        form_for: str | None = None,
        multi_line: bool | None = None,
        control_implementation: str | None = None,
        disabled: bool | None = None,
        printable: bool | None = None,
        xml_id: str | None = None,
        xforms_bind: str | None = None,
        form_id: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Create a FormFixedText, "form:fixed-text".

        The "form:fixed-text" element is usable within the following element:
        "form:form".

        Args:
            name: The name of the control.
            label: The label of the control.
            title: The title or tooltip of the control.
            form_for: The ID of the control this label is for.
            multi_line: If True, the label can have multiple lines.
            control_implementation: The control implementation.
            disabled: If True, the control is disabled.
            printable: If True, the control is printable.
            xml_id: The unique XML ID.
            xforms_bind: The XForms bind expression.
            form_id: The form ID (deprecated).
        """
        super().__init__(
            name=name,
            control_implementation=control_implementation,
            xml_id=xml_id,
            xforms_bind=xforms_bind,
            form_id=form_id,
            **kwargs,
        )
        if self._do_init:
            if label is not None:
                self.label = label
            if title is not None:
                self.title = title
            if form_for is not None:
                self.form_for = form_for
            if multi_line is not None:
                self.multi_line = multi_line
            if disabled is not None:
                self.disabled = disabled
            if printable is not None:
                self.printable = printable


FormFixedText._define_attribut_property()


class FormCombobox(FormSourceListMixin, FormSizetMixin, FormText):
    """A control which allows displaying and editing of text, and contains
    a list of possible values for that text, "form:combobox".

    Attributes:
        name (str or None): The name of the control (form:name).
        value (str or None): The value of the control (form:value).
        control_implementation (str or None): The control implementation (form:control-implementation).
        title (str or None): The title or tooltip of the control (form:title).
        disabled (bool): Specifies if the control is disabled (form:disabled).
        printable (bool): Specifies if the control is printable (form:printable).
        auto_complete (str or None): Specifies if auto-completion is enabled (form:auto-complete).
        dropdown (bool): Specifies if the combobox has a dropdown list (form:dropdown).
        list_source (str or None): The source of the list (form:list-source).
        source_cell_range (str or None): The cell range for the list source (form:source-cell-range).
        tab_stop (bool): Specifies if the control is a tab stop (form:tab-stop).
        readonly (bool): Specifies if the control is read-only (form:readonly).
        convert_empty_to_null (bool): Specifies if an empty value is converted to null (form:convert-empty-to-null).
        current_value (str or None): The current value of the control (form:current-value).
        data_field (str or None): The data field used for storing the value (form:data-field).
        linked_cell (str or None): The linked spreadsheet cell (form:linked-cell).
        xml_id (str or None): The unique XML ID (xml:id).
        xforms_bind (str or None): The XForms bind expression (xforms:bind).
        form_id (str or None): The form ID (deprecated, form:id).
        size (int or None): The number of visible items in the control (form:size).
        max_length (int or None): The maximum number of characters allowed (form:max-length).
        list_source_type (str or None): The type of the list source.
    """

    _tag = "form:combobox"
    _properties: tuple[PropDef | PropDefBool, ...] = (
        PropDef("name", "form:name"),
        PropDef("value", "form:value"),
        PropDef("control_implementation", "form:control-implementation"),
        PropDef("title", "form:title"),
        PropDefBool("disabled", "form:disabled", False),
        PropDefBool("printable", "form:printable", True),
        PropDef("auto_complete", "form:auto-complete"),
        PropDefBool("dropdown", "form:dropdown", False),
        PropDef("list_source", "form:list-source"),
        PropDef("source_cell_range", "form:source-cell-range"),
        PropDefBool("tab_stop", "form:tab-stop", True),
        PropDefBool("readonly", "form:readonly", False),
        PropDefBool("convert_empty_to_null", "form:convert-empty-to-null", False),
        PropDef("current_value", "form:current-value"),
        PropDef("data_field", "form:data-field"),
        PropDef("linked_cell", "form:linked-cell"),
        PropDef("xml_id", "xml:id"),
        PropDef("xforms_bind", "xforms:bind"),
        PropDef("form_id", "form:id"),  # deprecated
    )

    def __init__(
        self,
        name: str | None = None,
        value: str | None = None,
        control_implementation: str | None = None,
        title: str | None = None,
        disabled: bool | None = None,
        printable: bool | None = None,
        auto_complete: bool | None = None,
        dropdown: bool | None = None,
        list_source: str | None = None,
        list_source_type: str | None = None,
        source_cell_range: str | None = None,
        size: int | None = None,
        tab_index: int | None = None,
        tab_stop: bool | None = None,
        readonly: bool | None = None,
        convert_empty_to_null: bool | None = None,
        current_value: str | None = None,
        data_field: str | None = None,
        linked_cell: str | None = None,
        max_length: int | None = None,
        xml_id: str | None = None,
        xforms_bind: str | None = None,
        form_id: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Create a FormCombobox, "form:combobox".

        The "form:combobox" element is usable within the following elements:
        "form:column" and "form:form".

        Args:
            name: The name of the control.
            value: The value of the control.
            control_implementation: The control implementation.
            title: The title or tooltip of the control.
            disabled: If True, the control is disabled.
            printable: If True, the control is printable.
            auto_complete: If True, auto-complete is enabled.
            dropdown: If True, the combobox is a dropdown.
            list_source: The source of the list.
            list_source_type: The type of the list source.
            source_cell_range: The cell range for the list source.
            size: The size of the combobox.
            tab_index: The tab order of the control.
            tab_stop: If True, the control is a tab stop.
            readonly: If True, the control is read-only.
            convert_empty_to_null: If True, an empty value is converted to null.
            current_value: The current value of the control.
            data_field: The data field for storing the value.
            linked_cell: The linked spreadsheet cell.
            max_length: The maximum number of characters allowed.
            xml_id: The unique XML ID.
            xforms_bind: The XForms bind expression.
            form_id: The form ID (deprecated).
        """
        super().__init__(
            name=name,
            value=value,
            control_implementation=control_implementation,
            title=title,
            disabled=disabled,
            printable=printable,
            tab_index=tab_index,
            tab_stop=tab_stop,
            readonly=readonly,
            convert_empty_to_null=convert_empty_to_null,
            current_value=current_value,
            data_field=data_field,
            linked_cell=linked_cell,
            max_length=max_length,
            xml_id=xml_id,
            xforms_bind=xforms_bind,
            form_id=form_id,
            **kwargs,
        )
        if self._do_init:
            if auto_complete is not None:
                self.auto_complete = auto_complete
            if dropdown is not None:
                self.dropdown = dropdown
            if list_source is not None:
                self.list_source = list_source
            if list_source_type is not None:
                self.list_source_type = list_source_type
            if source_cell_range is not None:
                self.source_cell_range = source_cell_range
            if size is not None:
                self.size = size


FormCombobox._define_attribut_property()


class FormItem(Element):
    """A list item for a FormCombobox control, "form:item".

    Attributes:
        label (str or None): The label of the item (form:label).
    """

    _tag = "form:item"
    _properties: tuple[PropDef | PropDefBool, ...] = (PropDef("label", "form:label"),)

    def __init__(
        self,
        label: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Create a FormItem, "form:item".

        The "form:item" element is usable within the following element:
        "form:combobox".

        Args:
            label: The label of the item.
        """
        super().__init__(**kwargs)
        if self._do_init:
            self.label = label or ""


FormItem._define_attribut_property()


class FormListbox(FormSourceListMixin, FormSizetMixin, FormGrid):
    """An input control that allows a user to select one or more items from
    a list, "form:listbox".

    It is an alternative representation for a group of radio buttons.

    Attributes:
        name (str or None): The name of the control (form:name).
        control_implementation (str or None): The control implementation (form:control-implementation).
        title (str or None): The title or tooltip of the control (form:title).
        disabled (bool): Specifies if the control is disabled (form:disabled).
        printable (bool): Specifies if the control is printable (form:printable).
        dropdown (bool): Specifies if the listbox is a dropdown (form:dropdown).
        bound_column (str or None): The bound column for the listbox (form:bound-column).
        multiple (bool): Specifies if multiple selections are allowed (form:multiple).
        list_source (str or None): The source of the list (form:list-source).
        source_cell_range (str or None): The cell range for the list source (form:source-cell-range).
        tab_stop (bool): Specifies if the control is a tab stop (form:tab-stop).
        data_field (str or None): The data field used for storing the value (form:data-field).
        linked_cell (str or None): The linked spreadsheet cell (form:linked-cell).
        xml_id (str or None): The unique XML ID (xml:id).
        xforms_bind (str or None): The XForms bind expression (xforms:bind).
        xforms_list_source (str or None): The XForms list source (form:xforms-list-source).
        form_id (str or None): The form ID (deprecated, form:id).
        list_linkage_type (str or None): The list linkage type (form:list-linkage-type).
        size (int or None): The size of the listbox (form:size).
        list_source_type (str or None): The type of the list source (form:list-source-type).
    """

    _tag = "form:listbox"
    _properties: tuple[PropDef | PropDefBool, ...] = (
        PropDef("name", "form:name"),
        PropDef("control_implementation", "form:control-implementation"),
        PropDef("title", "form:title"),
        PropDefBool("disabled", "form:disabled", False),
        PropDefBool("printable", "form:printable", True),
        PropDefBool("dropdown", "form:dropdown", False),
        PropDef("bound_column", "form:bound-column"),
        PropDefBool("multiple", "form:multiple", False),
        PropDef("list_source", "form:list-source"),
        PropDef("source_cell_range", "form:source-cell-range"),
        PropDefBool("tab_stop", "form:tab-stop", True),
        PropDef("data_field", "form:data-field"),
        PropDef("linked_cell", "form:linked-cell"),
        PropDef("xml_id", "xml:id"),
        PropDef("xforms_bind", "xforms:bind"),
        PropDef("xforms_list_source", "form:xforms-list-source"),
        PropDef("form_id", "form:id"),  # deprecated
    )

    LIST_LINKAGE_TYPE: ClassVar[set[str]] = {
        "selection",
        "selection-indices",
    }

    def __init__(
        self,
        name: str | None = None,
        control_implementation: str | None = None,
        title: str | None = None,
        disabled: bool | None = None,
        printable: bool | None = None,
        dropdown: bool | None = None,
        bound_column: str | None = None,
        list_linkage_type: str | None = None,
        multiple: bool | None = None,
        list_source: str | None = None,
        list_source_type: str | None = None,
        source_cell_range: str | None = None,
        size: int | None = None,
        tab_index: int | None = None,
        tab_stop: bool | None = None,
        data_field: str | None = None,
        linked_cell: str | None = None,
        xml_id: str | None = None,
        xforms_bind: str | None = None,
        xforms_list_source: str | None = None,
        form_id: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Create a FormListbox, "form:listbox".

        The "form:listbox" element is usable within the following elements:
        "form:column" and "form:form".

        Args:
            name: The name of the control.
            control_implementation: The control implementation.
            title: The title or tooltip of the control.
            disabled: If True, the control is disabled.
            printable: If True, the control is printable.
            dropdown: If True, the listbox is a dropdown.
            bound_column: The bound column for the listbox.
            list_linkage_type: The list linkage type.
            multiple: If True, multiple selections are allowed.
            list_source: The source of the list.
            list_source_type: The type of the list source.
            source_cell_range: The cell range for the list source.
            size: The size of the listbox.
            tab_index: The tab order of the control.
            tab_stop: If True, the control is a tab stop.
            data_field: The data field for storing the value.
            linked_cell: The linked spreadsheet cell.
            xml_id: The unique XML ID.
            xforms_bind: The XForms bind expression.
            xforms_list_source: The XForms list source.
            form_id: The form ID (deprecated).
        """
        super().__init__(
            name=name,
            control_implementation=control_implementation,
            title=title,
            disabled=disabled,
            printable=printable,
            tab_index=tab_index,
            tab_stop=tab_stop,
            xml_id=xml_id,
            xforms_bind=xforms_bind,
            form_id=form_id,
            **kwargs,
        )
        if self._do_init:
            if data_field is not None:
                self.data_field = data_field
            if linked_cell is not None:
                self.linked_cell = linked_cell
            if dropdown is not None:
                self.dropdown = dropdown
            if bound_column is not None:
                self.bound_column = bound_column
            if list_linkage_type is not None:
                self.list_linkage_type = list_linkage_type
            if multiple is not None:
                self.multiple = multiple
            if list_source is not None:
                self.list_source = list_source
            if list_source_type is not None:
                self.list_source_type = list_source_type
            if source_cell_range is not None:
                self.source_cell_range = source_cell_range
            if size is not None:
                self.size = size
            if xforms_list_source is not None:
                self.xforms_list_source = xforms_list_source

    @property
    def list_linkage_type(self) -> str | None:
        return self.get_attribute_string("form:list-linkage-type")

    @list_linkage_type.setter
    def list_linkage_type(self, list_linkage_type: str | None) -> None:
        if list_linkage_type is None:
            self.del_attribute("form:list-linkage-type")
            return
        if list_linkage_type not in self.LIST_LINKAGE_TYPE:
            raise ValueError
        self.set_attribute("form:list-linkage-type", list_linkage_type)


FormListbox._define_attribut_property()


class FormOption(Element):
    """A list item for a "form:option" control, "form:option"."""

    _tag = "form:option"
    _properties: tuple[PropDef | PropDefBool, ...] = (
        PropDef("label", "form:label"),
        PropDef("value", "form:value"),
        PropDefBool("selected", "form:selected", False),
        PropDefBool("current_selected", "form:current-selected", False),
    )

    def __init__(
        self,
        label: str | None = None,
        value: str | None = None,
        selected: bool | None = None,
        current_selected: bool | None = None,
        **kwargs: Any,
    ) -> None:
        """Create a FormOption, "form:option".

        The "form:option" element is usable within the following element:
        "form:listbox".

        Args:
            label: The label of the option.
            value: The value of the option.
            selected: If True, the option is selected by default.
            current_selected: If True, the option is currently selected.
        """
        super().__init__(**kwargs)
        if self._do_init:
            self.label = label or ""
            if value is not None:
                self.value = value
            if selected is not None:
                self.selected = selected
            if current_selected is not None:
                self.current_selected = current_selected


FormOption._define_attribut_property()


class FormButton(
    OfficeTargetFrameMixin,
    FormDelayRepeatMixin,
    FormImageAlignMixin,
    FormImagePositionMixin,
    FormButtonTypeMixin,
    FormGrid,
):
    """A button control, "form:button"."""

    _tag = "form:button"
    _properties: tuple[PropDef | PropDefBool, ...] = (
        PropDef("name", "form:name"),
        PropDefBool("default_button", "form:default-button", False),
        PropDef("repeat", "form:repeat"),
        PropDefBool("disabled", "form:disabled", False),
        PropDef("focus_on_click", "form:focus-on-click"),
        PropDef("image_data", "form:image-data"),
        PropDef("label", "form:label"),
        PropDefBool("toggle", "form:toggle", False),
        PropDef("value", "form:value"),
        PropDef("href", "xlink:href"),
        PropDef("control_implementation", "form:control-implementation"),
        PropDef("title", "form:title"),
        PropDefBool("printable", "form:printable", True),
        PropDefBool("tab_stop", "form:tab-stop", True),
        PropDef("xml_id", "xml:id"),
        PropDef("xforms_bind", "xforms:bind"),
        PropDef("xforms_submission", "form:xforms-submission"),
        PropDef("form_id", "form:id"),  # deprecated
    )

    def __init__(
        self,
        name: str | None = None,
        title: str | None = None,
        label: str | None = None,
        button_type: str | None = None,
        default_button: bool | None = None,
        control_implementation: str | None = None,
        repeat: bool | None = None,
        delay_for_repeat: str | None = None,
        disabled: bool | None = None,
        focus_on_click: bool | None = None,
        image_align: str | None = None,
        image_position: str | None = None,
        image_data: str | None = None,
        value: str | None = None,
        toggle: bool | None = None,
        target_frame: str | None = None,
        href: str | None = None,
        printable: bool | None = None,
        tab_index: int | None = None,
        tab_stop: bool | None = None,
        xml_id: str | None = None,
        xforms_bind: str | None = None,
        xforms_submission: str | None = None,
        form_id: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Create a FormButton, "form:button".

        The "form:button" element is usable within the following element:
        "form:form".

        Args:
            name: The name of the control.
            value: The value of the control.
            control_implementation: The control implementation.
            title: The title or tooltip of the control.
            disabled: If True, the control is disabled.
            printable: If True, the control is printable.
            repeat: If True, the control repeats when a button is clicked.
            delay_for_repeat: The delay for repeating the action. Defaults to "PT0.050S".
            tab_index: The tab order of the control.
            tab_stop: If True, the control is a tab stop.
            button_type: The type of the button.
            default_button: If True, this is the default button.
            focus_on_click: If True, the button gets the focus when clicked.
            image_align: The alignment of the image.
            image_data: The image data for the button.
            image_position: The position of the image.
            label: The label of the button.
            toggle: If True, the button is a toggle button.
            target_frame: The target frame for the URL.
            href: The URL to navigate to when the button is clicked.
            xml_id: The unique XML ID.
            xforms_submission: The XForms submission to use.
            xforms_bind: The XForms bind expression.
            form_id: The form ID (deprecated).
        """
        super().__init__(
            name=name,
            control_implementation=control_implementation,
            disabled=disabled,
            printable=printable,
            title=title,
            tab_index=tab_index,
            tab_stop=tab_stop,
            xml_id=xml_id,
            xforms_bind=xforms_bind,
            form_id=form_id,
            **kwargs,
        )
        if self._do_init:
            if value is not None:
                self.value = value
            if button_type is not None:
                self.button_type = button_type
            if default_button is not None:
                self.default_button = default_button
            if repeat is not None:
                self.repeat = repeat
            if delay_for_repeat is not None:
                self.delay_for_repeat = delay_for_repeat
            if focus_on_click is not None:
                self.focus_on_click = focus_on_click
            if image_align is not None:
                self.image_align = image_align
            if image_data is not None:
                self.image_data = image_data
            if image_position is not None:
                self.image_position = image_position
            if label is not None:
                self.label = label
            if toggle is not None:
                self.toggle = toggle
            if xforms_submission is not None:
                self.xforms_submission = xforms_submission
            if target_frame is not None:
                self.target_frame = target_frame
            if href is not None:
                self.href = href


FormButton._define_attribut_property()


class FormImage(OfficeTargetFrameMixin, FormButtonTypeMixin, FormGrid):
    """A graphical button control, "form:image".

    Note: HTML 4.01 only allows the button type to be “submit” for an image
    button. In OpenDocument, an image button can be of any type.

    Attributes:
        button_type (str or None): The type of the button.
        control_implementation (str or None): The control implementation.
        disabled (bool): If True, the control is disabled.
        image_data (str or None): The image data for the button.
        name (str or None): The name of the control.
        printable (bool): If True, the control is printable.
        tab_index (int or None): The tab order of the control.
        tab_stop (bool): If True, the control is a tab stop.
        title (str): The title or tooltip of the control.
        value (str): The value of the control.
        target_frame (str): The target frame for the URL.
        xforms_bind (str): The XForms bind expression.
        href (str): The URL to navigate to when the button is clicked.
        xml_id (str): The unique XML ID.
        form_id (str): The form ID (deprecated).
    """

    _tag = "form:image"
    _properties: tuple[PropDef | PropDefBool, ...] = (
        PropDef("control_implementation", "form:control-implementation"),
        PropDefBool("disabled", "form:disabled", False),
        PropDef("image_data", "form:image-data"),
        PropDef("name", "form:name"),
        PropDefBool("printable", "form:printable", True),
        PropDef("tab_index", "form:tab-index"),
        PropDefBool("tab_stop", "form:tab-stop", True),
        PropDef("title", "form:title"),
        PropDef("value", "form:value"),
        PropDef("xforms_bind", "xforms:bind"),
        PropDef("href", "xlink:href"),
        PropDef("xml_id", "xml:id"),
        PropDef("form_id", "form:id"),  # deprecated
    )

    def __init__(
        self,
        name: str | None = None,
        title: str | None = None,
        value: str | None = None,
        button_type: str | None = None,
        control_implementation: str | None = None,
        disabled: bool | None = None,
        image_data: str | None = None,
        target_frame: str | None = None,
        href: str | None = None,
        printable: bool | None = None,
        tab_index: int | None = None,
        tab_stop: bool | None = None,
        xml_id: str | None = None,
        xforms_bind: str | None = None,
        form_id: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Create a FormImage, "form:image".

        The "form:image" element is usable within the following element:
        "form:form".

        Args:
            name: The name of the control.
            title: The title or tooltip of the control.
            value: The value of the control.
            button_type: The type of the button.
            control_implementation: The control implementation.
            disabled: If True, the control is disabled.
            image_data: The image data for the button.
            target_frame: The target frame for the URL.
            href: The URL to navigate to when the button is clicked.
            printable: If True, the control is printable.
            tab_index: The tab order of the control.
            tab_stop: If True, the control is a tab stop.
            xml_id: The unique XML ID.
            xforms_bind: The XForms bind expression.
            form_id: The form ID (deprecated).
        """
        super().__init__(
            name=name,
            control_implementation=control_implementation,
            disabled=disabled,
            printable=printable,
            title=title,
            tab_index=tab_index,
            tab_stop=tab_stop,
            xml_id=xml_id,
            xforms_bind=xforms_bind,
            form_id=form_id,
            **kwargs,
        )
        if self._do_init:
            if value is not None:
                self.value = value
            if button_type is not None:
                self.button_type = button_type
            if image_data is not None:
                self.image_data = image_data
            if target_frame is not None:
                self.target_frame = target_frame
            if href is not None:
                self.href = href


FormImage._define_attribut_property()


class FormCheckbox(FormImageAlignMixin, FormImagePositionMixin, FormGrid):
    """An on/off control, "form:checkbox".

    The control is on when the value of the form:current-state attribute
    associated with the control element is checked.

    Attributes:
        control_implementation (str or None): The control implementation.
        current_state (str or None): The current state of the checkbox.
        data_field (str or None): The data field used for storing the value.
        disabled (bool): If True, the control is disabled.
        id (str or None): The unique XML ID (form:id).
        image_align (str or None): The alignment of the image.
        image_position (str or None): The position of the image.
        is_tristate (bool): If True, the checkbox can have three states.
        label (str or None): The label of the control.
        linked_cell (str or None): The linked spreadsheet cell.
        name (str or None): The name of the control.
        printable (bool): If True, the control is printable.
        state (str or None): The state of the checkbox.
        tab_index (int or None): The tab order of the control.
        tab_stop (bool): If True, the control is a tab stop.
        title (str or None): The title or tooltip of the control.
        value (str or None): The value of the control.
        visual_effect (str or None): The visual effect of the control.
        xforms_bind (str or None): The XForms bind expression.
        xml_id (str or None): The unique XML ID.
        form_id (str or None): The form ID (deprecated).
    """

    _tag = "form:checkbox"
    _properties: tuple[PropDef | PropDefBool, ...] = (
        PropDef("control_implementation", "form:control-implementation"),
        PropDef("current_state", "form:current-state"),
        PropDef("data_field", "form:data-field"),
        PropDefBool("disabled", "form:disabled", False),
        PropDef("image_align", "form:image-align"),
        PropDef("image_position", "form:image-position"),
        PropDefBool("is_tristate", "form:is-tristate", False),
        PropDef("label", "form:label"),
        PropDef("linked_cell", "form:linked-cell"),
        PropDef("name", "form:name"),
        PropDefBool("printable", "form:printable", True),
        PropDef("state", "form:state"),
        PropDef("tab_index", "form:tab-index"),
        PropDefBool("tab_stop", "form:tab-stop", True),
        PropDef("title", "form:title"),
        PropDef("value", "form:value"),
        PropDef("visual_effect", "form:visual-effect"),
        PropDef("xforms_bind", "xforms:bind"),
        PropDef("xml_id", "xml:id"),
        PropDef("form_id", "form:id"),  # deprecated
    )

    def __init__(
        self,
        name: str | None = None,
        title: str | None = None,
        value: str | None = None,
        control_implementation: str | None = None,
        current_state: str | None = None,
        data_field: str | None = None,
        disabled: bool | None = None,
        image_align: str | None = None,
        image_position: str | None = None,
        is_tristate: bool | None = None,
        label: str | None = None,
        linked_cell: str | None = None,
        printable: bool | None = None,
        state: str | None = None,
        tab_index: int | None = None,
        tab_stop: bool | None = None,
        visual_effect: str | None = None,
        xml_id: str | None = None,
        xforms_bind: str | None = None,
        form_id: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Create a FormCheckbox, "form:checkbox".

        The "form:checkbox" element is usable within the following elements:
        "form:column" and "form:form".

        Args:
            name: The name of the control.
            title: The title or tooltip of the control.
            value: The value of the control.
            control_implementation: The control implementation.
            current_state: The current state of the checkbox.
            data_field: The data field used for storing the value.
            disabled: If True, the control is disabled.
            image_align: The alignment of the image.
            image_position: The position of the image.
            is_tristate: If True, the checkbox can have three states.
            label: The label of the control.
            linked_cell: The linked spreadsheet cell.
            printable: If True, the control is printable.
            state: The state of the checkbox.
            tab_index: The tab order of the control.
            tab_stop: If True, the control is a tab stop.
            visual_effect: The visual effect of the control.
            xml_id: The unique XML ID.
            xforms_bind: The XForms bind expression.
            form_id: The form ID (deprecated).
        """
        super().__init__(
            name=name,
            control_implementation=control_implementation,
            disabled=disabled,
            printable=printable,
            title=title,
            tab_index=tab_index,
            tab_stop=tab_stop,
            xml_id=xml_id,
            xforms_bind=xforms_bind,
            form_id=form_id,
            **kwargs,
        )
        if self._do_init:
            if current_state is not None:
                self.current_state = current_state
            if data_field is not None:
                self.data_field = data_field
            if image_align is not None:
                self.image_align = image_align
            if image_position is not None:
                self.image_position = image_position
            if is_tristate is not None:
                self.is_tristate = is_tristate
            if label is not None:
                self.label = label
            if linked_cell is not None:
                self.linked_cell = linked_cell
            if state is not None:
                self.state = state
            if value is not None:
                self.value = value
            if visual_effect is not None:
                self.visual_effect = visual_effect


FormCheckbox._define_attribut_property()


class FormRadio(FormImageAlignMixin, FormImagePositionMixin, FormGrid):
    """A control which acts like a check box except that when multiple radio
    buttons belong to the same group they are mutually exclusive, "form:radio".

    Radio buttons are defined to belong to the same group if they have the same
    control name, as specified by their form:name attribute.

    If a group of radio buttons is bound to one database field, and a user
    selects any given button, the reference value of the selected radio button
    is written into its database field.

    Attributes:
        control_implementation (str or None): The control implementation.
        current_selected (bool): If True, the radio button is currently selected.
        data_field (str or None): The data field used for storing the value.
        disabled (bool): If True, the control is disabled.
        id (str or None): The unique XML ID (form:id).
        image_align (str or None): The alignment of the image.
        image_position (str or None): The position of the image.
        label (str or None): The label of the control.
        linked_cell (str or None): The linked spreadsheet cell.
        name (str or None): The name of the control.
        printable (bool): If True, the control is printable.
        selected (bool): If True, the radio button is selected by default.
        tab_index (int or None): The tab order of the control.
        tab_stop (bool): If True, the control is a tab stop.
        title (str or None): The title or tooltip of the control.
        value (str or None): The value of the control.
        visual_effect (str or None): The visual effect of the control.
        xforms_bind (str or None): The XForms bind expression.
        xml_id (str or None): The unique XML ID.
        form_id (str or None): The form ID (deprecated).
    """

    _tag = "form:radio"
    _properties: tuple[PropDef | PropDefBool, ...] = (
        PropDef("control_implementation", "form:control-implementation"),
        PropDefBool("current_selected", "form:current-selected", False),
        PropDef("data_field", "form:data-field"),
        PropDefBool("disabled", "form:disabled", False),
        PropDef("image_align", "form:image-align"),
        PropDef("image_position", "form:image-position"),
        PropDef("label", "form:label"),
        PropDef("linked_cell", "form:linked-cell"),
        PropDef("name", "form:name"),
        PropDefBool("printable", "form:printable", True),
        PropDefBool("selected", "form:selected", False),
        PropDef("tab_index", "form:tab-index"),
        PropDefBool("tab_stop", "form:tab-stop", True),
        PropDef("title", "form:title"),
        PropDef("value", "form:value"),
        PropDef("visual_effect", "form:visual-effect"),
        PropDef("xforms_bind", "xforms:bind"),
        PropDef("xml_id", "xml:id"),
        PropDef("form_id", "form:id"),  # deprecated
    )

    def __init__(
        self,
        name: str | None = None,
        title: str | None = None,
        value: str | None = None,
        control_implementation: str | None = None,
        current_selected: bool | None = None,
        data_field: str | None = None,
        disabled: bool | None = None,
        image_align: str | None = None,
        image_position: str | None = None,
        label: str | None = None,
        linked_cell: str | None = None,
        printable: bool | None = None,
        selected: bool | None = None,
        tab_index: int | None = None,
        tab_stop: bool | None = None,
        visual_effect: str | None = None,
        xml_id: str | None = None,
        xforms_bind: str | None = None,
        form_id: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Create a FormRadio, "form:radio".

        The "form:radio" element is usable within the following element:
        "form:form".

        Args:
            name: The name of the control.
            title: The title or tooltip of the control.
            value: The value of the control.
            control_implementation: The control implementation.
            current_selected: If True, the radio button is currently selected.
            data_field: The data field used for storing the value.
            disabled: If True, the control is disabled.
            image_align: The alignment of the image.
            image_position: The position of the image.
            label: The label of the control.
            linked_cell: The linked spreadsheet cell.
            printable: If True, the control is printable.
            selected: If True, the radio button is selected by default.
            tab_index: The tab order of the control.
            tab_stop: If True, the control is a tab stop.
            visual_effect: The visual effect of the control.
            xml_id: The unique XML ID.
            xforms_bind: The XForms bind expression.
            form_id: The form ID (deprecated).
        """
        super().__init__(
            name=name,
            control_implementation=control_implementation,
            disabled=disabled,
            printable=printable,
            title=title,
            tab_index=tab_index,
            tab_stop=tab_stop,
            xml_id=xml_id,
            xforms_bind=xforms_bind,
            form_id=form_id,
            **kwargs,
        )
        if self._do_init:
            if current_selected is not None:
                self.current_selected = current_selected
            if data_field is not None:
                self.data_field = data_field
            if image_align is not None:
                self.image_align = image_align
            if image_position is not None:
                self.image_position = image_position
            if label is not None:
                self.label = label
            if linked_cell is not None:
                self.linked_cell = linked_cell
            if selected is not None:
                self.selected = selected
            if value is not None:
                self.value = value
            if visual_effect is not None:
                self.visual_effect = visual_effect


FormRadio._define_attribut_property()


class FormFrame(FormGenericControl):
    """A frame in which controls may be visually arranged, "form:frame".

    Attributes:
        control_implementation (str or None): The control implementation.
        disabled (bool): If True, the control is disabled.
        form_for (str or None): The ID of the control this frame is for.
        id (str or None): The unique XML ID (form:id).
        label (str or None): The label of the control.
        name (str or None): The name of the control.
        printable (bool): If True, the control is printable.
        title (str or None): The title or tooltip of the control.
        xforms_bind (str or None): The XForms bind expression.
        xml_id (str or None): The unique XML ID.
        form_id (str or None): The form ID (deprecated).
    """

    _tag = "form:frame"
    _properties: tuple[PropDef | PropDefBool, ...] = (
        PropDef("control_implementation", "form:control-implementation"),
        PropDefBool("disabled", "form:disabled", False),
        PropDef("form_for", "form:for"),
        PropDef("label", "form:label"),
        PropDef("name", "form:name"),
        PropDefBool("printable", "form:printable", True),
        PropDef("title", "form:title"),
        PropDef("xforms_bind", "xforms:bind"),
        PropDef("xml_id", "xml:id"),
        PropDef("form_id", "form:id"),  # deprecated
    )

    def __init__(
        self,
        name: str | None = None,
        title: str | None = None,
        label: str | None = None,
        form_for: str | None = None,
        control_implementation: str | None = None,
        disabled: bool | None = None,
        printable: bool | None = None,
        xml_id: str | None = None,
        xforms_bind: str | None = None,
        form_id: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Create a FormFrame, "form:frame".

        The "form:frame" element is usable within the following element:
        "form:form".

        Args:
            name: The name of the control.
            title: The title or tooltip of the control.
            label: The label of the control.
            form_for: The ID of the control this frame is for.
            control_implementation: The control implementation.
            disabled: If True, the control is disabled.
            printable: If True, the control is printable.
            xml_id: The unique XML ID.
            xforms_bind: The XForms bind expression.
            form_id: The form ID (deprecated).
        """
        super().__init__(
            name=name,
            control_implementation=control_implementation,
            xml_id=xml_id,
            xforms_bind=xforms_bind,
            form_id=form_id,
            **kwargs,
        )
        if self._do_init:
            if disabled is not None:
                self.disabled = disabled
            if form_for is not None:
                self.form_for = form_for
            if label is not None:
                self.label = label
            if printable is not None:
                self.printable = printable
            if title is not None:
                self.title = title


FormFrame._define_attribut_property()


class FormImageFrame(FormGenericControl):
    """A graphical control displaying an image, "form:image-frame".

    The control displays an image, whose location is described in the control.

    Attributes:
        control_implementation (str or None): The control implementation.
        data_field (str or None): The data field used for storing the value.
        disabled (bool): If True, the control is disabled.
        id (str or None): The unique XML ID (form:id).
        image_data (str or None): The image data for the control.
        name (str or None): The name of the control.
        printable (bool): If True, the control is printable.
        readonly (bool): If True, the control is read-only.
        title (str or None): The title or tooltip of the control.
        xforms_bind (str or None): The XForms bind expression.
        xml_id (str or None): The unique XML ID.
        form_id (str or None): The form ID (deprecated).
    """

    _tag = "form:image-frame"
    _properties: tuple[PropDef | PropDefBool, ...] = (
        PropDef("control_implementation", "form:control-implementation"),
        PropDef("data_field", "form:data-field"),
        PropDefBool("disabled", "form:disabled", False),
        PropDef("image_data", "form:image-data"),
        PropDef("name", "form:name"),
        PropDefBool("printable", "form:printable", True),
        PropDefBool("readonly", "form:readonly", False),
        PropDef("title", "form:title"),
        PropDef("xforms_bind", "xforms:bind"),
        PropDef("xml_id", "xml:id"),
        PropDef("form_id", "form:id"),  # deprecated
    )

    def __init__(
        self,
        name: str | None = None,
        title: str | None = None,
        control_implementation: str | None = None,
        data_field: str | None = None,
        disabled: bool | None = None,
        image_data: str | None = None,
        printable: bool | None = None,
        readonly: bool | None = None,
        xml_id: str | None = None,
        xforms_bind: str | None = None,
        form_id: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Create a FormImageFrame, "form:image-frame".

        The "form:image-frame" element is usable within the following element:
        "form:form".

        Args:
            name: The name of the control.
            title: The title or tooltip of the control.
            control_implementation: The control implementation.
            data_field: The data field used for storing the value.
            disabled: If True, the control is disabled.
            image_data: The image data for the control.
            printable: If True, the control is printable.
            readonly: If True, the control is read-only.
            xml_id: The unique XML ID.
            xforms_bind: The XForms bind expression.
            form_id: The form ID (deprecated).
        """
        super().__init__(
            name=name,
            control_implementation=control_implementation,
            xml_id=xml_id,
            xforms_bind=xforms_bind,
            form_id=form_id,
            **kwargs,
        )
        if self._do_init:
            if data_field is not None:
                self.data_field = data_field
            if disabled is not None:
                self.disabled = disabled
            if image_data is not None:
                self.image_data = image_data
            if printable is not None:
                self.printable = printable
            if readonly is not None:
                self.readonly = readonly
            if title is not None:
                self.title = title


FormImageFrame._define_attribut_property()


class FormValueRange(FormDelayRepeatMixin, FormGrid):
    """A control which allows the user to select a value from a number range,
    "form:value-range".

    Attributes:
        control_implementation (str or None): The control implementation.
        delay_for_repeat (str or None): The delay for repeating the action.
        disabled (bool): If True, the control is disabled.
        id (str or None): The unique XML ID (form:id).
        linked_cell (str or None): The linked spreadsheet cell.
        max_value (str or None): The maximum value allowed.
        min_value (str or None): The minimum value allowed.
        name (str or None): The name of the control.
        orientation (str or None): The orientation of the control.
        page_step_size (str or None): The page step size.
        printable (bool): If True, the control is printable.
        repeat (bool): If True, the control repeats when a button is clicked.
        step_size (str or None): The step size.
        tab_index (int or None): The tab order of the control.
        tab_stop (bool): If True, the control is a tab stop.
        title (str or None): The title or tooltip of the control.
        value (str or None): The value of the control.
        xforms_bind (str or None): The XForms bind expression.
        xml_id (str or None): The unique XML ID.
        form_id (str or None): The form ID (deprecated).
    """

    _tag = "form:value-range"
    _properties: tuple[PropDef | PropDefBool, ...] = (
        PropDef("control_implementation", "form:control-implementation"),
        PropDef("delay_for_repeat", "form:delay-for-repeat"),
        PropDefBool("disabled", "form:disabled", False),
        PropDef("linked_cell", "form:linked-cell"),
        PropDef("max_value", "form:max-value"),
        PropDef("min_value", "form:min-value"),
        PropDef("name", "form:name"),
        PropDef("orientation", "form:orientation"),
        PropDef("page_step_size", "form:page-step-size"),
        PropDefBool("printable", "form:printable", True),
        PropDef("repeat", "form:repeat"),
        PropDef("step_size", "form:step-size"),
        PropDefBool("tab_stop", "form:tab-stop", True),
        PropDef("title", "form:title"),
        PropDef("value", "form:value"),
        PropDef("xforms_bind", "xforms:bind"),
        PropDef("xml_id", "xml:id"),
        PropDef("form_id", "form:id"),  # deprecated
    )

    def __init__(
        self,
        name: str | None = None,
        title: str | None = None,
        value: str | None = None,
        control_implementation: str | None = None,
        delay_for_repeat: str | None = None,
        disabled: bool | None = None,
        linked_cell: str | None = None,
        max_value: str | None = None,
        min_value: str | None = None,
        orientation: str | None = None,
        page_step_size: str | None = None,
        printable: bool | None = None,
        repeat: bool | None = None,
        step_size: str | None = None,
        tab_index: int | None = None,
        tab_stop: bool | None = None,
        xml_id: str | None = None,
        xforms_bind: str | None = None,
        form_id: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Create a FormValueRange, "form:value-range".

        The "form:value-range" element is usable within the following element:
        "form:form".

        Args:
            name: The name of the control.
            title: The title or tooltip of the control.
            value: The value of the control.
            control_implementation: The control implementation.
            delay_for_repeat: The delay for repeating the action.
            disabled: If True, the control is disabled.
            linked_cell: The linked spreadsheet cell.
            max_value: The maximum value allowed.
            min_value: The minimum value allowed.
            orientation: The orientation of the control.
            page_step_size: The page step size.
            printable: If True, the control is printable.
            repeat: If True, the control repeats when a button is clicked.
            step_size: The step size.
            tab_index: The tab order of the control.
            tab_stop: If True, the control is a tab stop.
            xml_id: The unique XML ID.
            xforms_bind: The XForms bind expression.
            form_id: The form ID (deprecated).
        """
        super().__init__(
            name=name,
            title=title,
            control_implementation=control_implementation,
            disabled=disabled,
            printable=printable,
            tab_index=tab_index,
            tab_stop=tab_stop,
            xml_id=xml_id,
            xforms_bind=xforms_bind,
            form_id=form_id,
            **kwargs,
        )
        if self._do_init:
            if delay_for_repeat is not None:
                self.delay_for_repeat = delay_for_repeat
            if linked_cell is not None:
                self.linked_cell = linked_cell
            if max_value is not None:
                self.max_value = max_value
            if min_value is not None:
                self.min_value = min_value
            if orientation is not None:
                self.orientation = orientation
            if page_step_size is not None:
                self.page_step_size = page_step_size
            if repeat is not None:
                self.repeat = repeat
            if step_size is not None:
                self.step_size = step_size
            if value is not None:
                self.value = value


FormValueRange._define_attribut_property()


register_element_class(FormButton)
register_element_class(FormCheckbox)
register_element_class(FormColumn)
register_element_class(FormCombobox)
register_element_class(FormDate)
register_element_class(FormFile)
register_element_class(FormFixedText)
register_element_class(FormFormattedText)
register_element_class(FormFrame)
register_element_class(FormGenericControl)
register_element_class(FormGrid)
register_element_class(FormHidden)
register_element_class(FormImage)
register_element_class(FormImageFrame)
register_element_class(FormItem)
register_element_class(FormListbox)
register_element_class(FormNumber)
register_element_class(FormOption)
register_element_class(FormPassword)
register_element_class(FormRadio)
register_element_class(FormText)
register_element_class(FormTextarea)
register_element_class(FormTime)
register_element_class(FormValueRange)



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2013 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
"""Body, root of the document content."""

from __future__ import annotations

from typing import Any, cast

from .annotation import AnnotationMixin
from .bookmark import BookmarkMixin
from .config_elements import _from_dict
from .const import BODY_ALLOW_NAMED_RANGE_TAGS
from .element import Element, PropDef, PropDefBool, register_element_class
from .form import FormMixin
from .mixin_link import LinkMixin
from .mixin_list import ListMixin
from .mixin_named_range import NRMixin
from .mixin_toc import TocMixin
from .note import NoteMixin
from .office_forms import OfficeFormsMixin
from .reference import ReferenceMixin
from .section import SectionMixin
from .table import Table
from .tracked_changes import TrackedChangesMixin
from .user_field import UserDefinedMixin
from .user_field_declaration import UserFieldDeclContMixin
from .variable_declaration import VarDeclMixin

# for compatibility with version <= 3.18.1
BODY_NR_TAGS = BODY_ALLOW_NAMED_RANGE_TAGS


class Body(Element):
    """Root of the document content, "office:body"."""

    _tag: str = "office:body"
    _properties: tuple[PropDef | PropDefBool, ...] = ()

    def get_tables(
        self,
        style: str | None = None,
        content: str | None = None,
    ) -> list[Table]:
        """Return all the tables that match the criteria.

        The method is also accessible via the alias `get_sheets()`.

        Args:
            style: The style name of the tables to match.
            content: A regex pattern to match within the table content.

        Returns:
            list[Table]: A list of matching Table elements.
        """
        return self._filtered_elements(  # type: ignore[return-value]
            "descendant::table:table",
            table_style=style,
            content=content,
        )

    get_sheets = get_tables

    @property
    def tables(self) -> list[Table]:
        """Return all the tables.

        The property is also accessible via the alias `sheets`.

        Returns:
            list[Table]: A list of all Table elements.
        """
        return cast(list[Table], self.get_elements("descendant::table:table"))

    sheets = tables

    def get_table(
        self,
        position: int = 0,
        name: str | None = None,
        content: str | None = None,
    ) -> Table | None:
        """Return the table that matches the criteria.

        The method is also accessible via the alias `get_sheet()`.

        Args:
            position: The 0-based index of the table to retrieve
                among the matching tables. Defaults to 0.
            name: The name of the table to match.
            content: A regex pattern to match within the table content.

        Returns:
            Table or None: The matching Table element, or None if not found.
        """
        if name is None and content is None:
            result = self._filtered_element("descendant::table:table", position)
        else:
            result = self._filtered_element(
                "descendant::table:table",
                position,
                table_name=name,
                content=content,
            )
        return result  # type: ignore[return-value]

    get_sheet = get_table

    @property
    def allow_named_range(self) -> bool:
        """Return True if the current body allows named ranges."""
        return self.tag in BODY_ALLOW_NAMED_RANGE_TAGS


class Chart(UserFieldDeclContMixin, VarDeclMixin, NRMixin, Body):
    """Root of the Chart document content, "office:chart"."""

    _tag: str = "office:chart"
    _properties: tuple[PropDef | PropDefBool, ...] = ()


class Database(Body):
    """Root of the Database document content, "office:database"."""

    _tag: str = "office:database"
    _properties: tuple[PropDef | PropDefBool, ...] = ()


class Drawing(UserFieldDeclContMixin, VarDeclMixin, NRMixin, Body):
    """Root of the Drawing document content, "office:drawing"."""

    _tag: str = "office:drawing"
    _properties: tuple[PropDef | PropDefBool, ...] = ()


class Image(Body):
    """Root of the Image document content, "office:image"."""

    _tag: str = "office:image"
    _properties: tuple[PropDef | PropDefBool, ...] = ()


class Presentation(
    UserDefinedMixin, UserFieldDeclContMixin, LinkMixin, VarDeclMixin, NRMixin, Body
):
    """Root of the Presentation document content, "office:presentation"."""

    _tag: str = "office:presentation"
    _properties: tuple[PropDef | PropDefBool, ...] = ()


class Spreadsheet(
    UserDefinedMixin,
    UserFieldDeclContMixin,
    VarDeclMixin,
    LinkMixin,
    AnnotationMixin,
    NRMixin,
    Body,
):
    """Root of the Spreadsheet document content, "office:spreadsheet"."""

    _tag: str = "office:spreadsheet"
    _properties: tuple[PropDef | PropDefBool, ...] = ()


class Text(
    ListMixin,
    TocMixin,
    UserDefinedMixin,
    UserFieldDeclContMixin,
    LinkMixin,
    VarDeclMixin,
    NRMixin,
    FormMixin,
    TrackedChangesMixin,
    OfficeFormsMixin,
    SectionMixin,
    ReferenceMixin,
    BookmarkMixin,
    AnnotationMixin,
    NoteMixin,
    Body,
):
    """Root of the Text document content, "office:text"."""

    _tag: str = "office:text"
    _properties: tuple[PropDef | PropDefBool, ...] = ()


class Metadata(Body):
    """Root of the Meta document content, "office:meta"."""

    _tag: str = "office:meta"
    _properties: tuple[PropDef | PropDefBool, ...] = ()


class OfficeSettings(Body):
    """Root of the Settings document content, "office:settings"."""

    _tag: str = "office:settings"
    _properties: tuple[PropDef | PropDefBool, ...] = ()

    def as_dict(self) -> dict[str, str | int | bool | dict[str, Any] | list[Any]]:
        """Serialize the OfficeSettings element and its children to a dictionary.

        This method recursively converts the element and its child elements
        (if they also have an `as_dict` method) into a dictionary
        representation.

        Returns:
            dict: A dictionary representing the OfficeSettings element,
                where the key is the element's tag and the value is a
                dictionary containing its children's representations.
        """
        conf: dict[str, str | int | bool | dict[str, Any] | list[Any]] = {
            "class": self._tag
        }
        children = [child.as_dict() for child in self.children]  # type: ignore[attr-defined]
        if children:
            conf["children"] = children
        return conf

    @classmethod
    def from_dict(
        cls, data: dict[str, str | int | bool | dict[str, Any]]
    ) -> OfficeSettings:
        """Create an OfficeSettings element from a dictionary representation.

        This class method reconstructs an OfficeSettings element and its
        children from a dictionary previously generated by a method like
        `as_dict()`. It leverages an internal helper `_from_dict` for the
        actual conversion.

        Args:
            data: A dictionary containing the representation of the
                OfficeSettings element, including its tag and children.

        Returns:
            An OfficeSettings instance populated with data from the dictionary.
        """
        return cast(OfficeSettings, _from_dict(data))


register_element_class(Body)
register_element_class(Chart)
register_element_class(Database)
register_element_class(Drawing)
register_element_class(Image)
register_element_class(Metadata)
register_element_class(OfficeSettings)
register_element_class(Presentation)
register_element_class(Spreadsheet)
register_element_class(Text)



# Copyright 2018-2026 Jérôme Dumonteil
# Copyright (c) 2009-2010 Ars Aperta, Itaapy, Pierlis, Talend.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Authors (odfdo project): jerome.dumonteil@gmail.com
# The odfdo project is a derivative work of the lpod-python project:
# https://github.com/lpod/lpod-python
# Authors: David Versmisse <david.versmisse@itaapy.com>

from collections.abc import Iterable

import pytest

from odfdo.document import Document
from odfdo.element import Element
from odfdo.toc import (
    TOC,
    IndexTitle,
    IndexTitleTemplate,
    TabStopStyle,
    TocEntryTemplate,
)

SAMPLE_EXPECTED = [
    "Table des matières",
    "1. Level 1 title 1",
    "1.1. Level 2 title 1",
    "2. Level 1 title 2",
    "2.1.1. Level 3 title 1",
    "2.2. Level 2 title 2",
    "3. Level 1 title 3",
    "3.1. Level 2 title 1",
    "3.1.1. Level 3 title 1",
    "3.1.2. Level 3 title 2",
    "3.2. Level 2 title 2",
    "3.2.1. Level 3 title 1",
    "3.2.2. Level 3 title 2",
]


@pytest.fixture
def sample(samples) -> Iterable[Document]:
    document = Document(samples("toc.odt"))
    yield document


@pytest.fixture
def sample_toc(samples) -> Iterable[Document]:
    document = Document(samples("toc_done.odt"))
    yield document


def get_toc_lines(toc):
    return [paragraph.text for paragraph in toc.paragraphs]


def test_index_title_class():
    title = IndexTitle()
    assert isinstance(title, IndexTitle)


def test_index_title_style():
    title = IndexTitle(style="Standard", xml_id="abc")
    assert isinstance(title, IndexTitle)


def test_tab_stop_style_class():
    element = TabStopStyle()
    assert isinstance(element, TabStopStyle)


def test_tab_stop_style_class_from_tag():
    element = Element.from_tag("<style:tab-stop/>")
    assert isinstance(element, TabStopStyle)


def test_index_title_title_text_1():
    title = IndexTitle(
        style="Standard",
        xml_id="abc",
        title_text="some title",
        title_text_style="some_style_name",
    )
    paragraph = title.get_paragraph()
    assert paragraph is not None
    assert str(paragraph).strip() == "some title"
    assert paragraph.style == "some_style_name"


def test_index_title_title_text_2():
    title = IndexTitle(
        style="Standard",
        xml_id="abc",
        title_text="some title",
        title_text_style="some_style_name",
    )
    title.set_title_text("new title", "new_style")
    paragraph = title.get_paragraph()
    assert paragraph is not None
    assert str(paragraph).strip() == "new title"
    assert paragraph.style == "new_style"


def test_tab_stop_style_argsg():
    element = TabStopStyle(
        style_type="right",
        leader_style="dotted",
        leader_text=".",
        style_char="Standard",
        leader_color="#000",
        leader_text_style="dotted",
        leader_type="Standard",
        leader_width="1cm",
        style_position="default",
    )
    assert isinstance(element, TabStopStyle)


def test_get_tocs(sample_toc):
    tocs = sample_toc.body.get_tocs()
    assert len(tocs) == 1
    assert isinstance(tocs[0], TOC)


def test_get_tocs_empty(sample):
    tocs = sample.body.get_tocs()
    assert len(tocs) == 0


def test_get_tocs_property(sample_toc):
    tocs = sample_toc.body.tocs
    assert len(tocs) == 1
    assert isinstance(tocs[0], TOC)


def test_get_toc(sample_toc):
    toc = sample_toc.body.get_toc()
    assert isinstance(toc, TOC)


def test_get_toc_property(sample_toc):
    toc = sample_toc.body.toc
    assert isinstance(toc, TOC)


def test_toc_fill_unattached():
    toc = TOC("Table des matières")
    with pytest.raises(ValueError):
        toc.fill()


def test_toc_fill_unattached_document(sample):
    toc = TOC("Table des matières")
    toc.fill(sample)
    toc_lines = get_toc_lines(toc)
    assert toc_lines == SAMPLE_EXPECTED


def test_toc_fill_unattached_document_no_title(sample):
    toc = TOC("")
    toc.fill(sample)
    toc_lines = get_toc_lines(toc)
    assert toc_lines == SAMPLE_EXPECTED[1:]


def test_toc_fill_unattached_document_no_style(sample):
    toc = TOC("Table des matières")
    toc.fill(sample, use_default_styles=False)
    toc_lines = get_toc_lines(toc)
    assert toc_lines == SAMPLE_EXPECTED


def test_toc_fill_unattached_document_outline_level(sample):
    toc = TOC("Table des matières", outline_level=1)
    toc.fill(sample)
    toc_lines = get_toc_lines(toc)
    expected = [
        "Table des matières",
        "1. Level 1 title 1",
        "2. Level 1 title 2",
        "3. Level 1 title 3",
    ]
    assert toc_lines == expected


def test_toc_fill_attached(sample):
    document = sample.clone
    toc = TOC("Table des matières")
    document.body.append(toc)
    toc.fill()
    toc_lines = get_toc_lines(toc)
    assert toc_lines == SAMPLE_EXPECTED


def test_toc_repr_empty():
    toc = TOC("Table des matières")
    assert repr(toc) == "<TOC tag=text:table-of-content>"


def test_toc_str_empty():
    toc = TOC("Table des matières")
    assert "Table des matières" in str(toc)


def test_toc_repr(sample):
    toc = TOC("Table des matières")
    toc.fill(sample)
    assert repr(toc) == "<TOC tag=text:table-of-content>"


def test_toc_str(sample):
    toc = TOC("Table des matières")
    toc.fill(sample)
    result = str(toc)
    for line in SAMPLE_EXPECTED:
        assert line in result


def test_toc_args():
    toc = TOC(
        title="Table des matières", name="table", style="Standard", protected=False
    )
    assert "Table des matières" in str(toc)


def test_toc_no_title_bad_args():
    toc = TOC(title="")
    assert "Table des matières" not in str(toc)


def test_toc_index_title_template_create():
    index_title = IndexTitleTemplate()
    assert index_title.tag == "text:index-title-template"


def test_toc_index_title_template_create_no_style():
    index_title = IndexTitleTemplate()
    assert index_title.style is None


def test_toc_index_title_template_create_style():
    index_title = IndexTitleTemplate(style="Standard")
    assert index_title.style == "Standard"


def test_toc_entry_template_create():
    entry = TocEntryTemplate()
    assert entry.tag == "text:table-of-content-entry-template"


def test_toc_entry_template_no_style():
    entry = TocEntryTemplate()
    assert entry.style is None


def test_toc_entry_template_style():
    entry = TocEntryTemplate(style="Standard")
    assert entry.style == "Standard"


def test_toc_entry_template_no_outline_level():
    entry = TocEntryTemplate()
    assert entry.outline_level is None


def test_toc_entry_template_outline_level():
    entry = TocEntryTemplate(outline_level=1)
    assert entry.outline_level == 1


def test_toc_entry_template_complete_defaults():
    entry = TocEntryTemplate(outline_level=1)
    entry.complete_defaults()
    expected = (
        "<text:table-of-content-entry-template "
        'text:outline-level="1">'
        "<text:index-entry-chapter/>"
        "<text:index-entry-text/>"
        "<text:index-entry-text/>"
        '<text:index-entry-text style:type="right" '
        'style:leader-char="."/>'
        "<text:index-entry-page-number/>"
        "</text:table-of-content-entry-template>"
    )
    assert entry.serialize() == expected


def test_toc_create_toc_source_args():
    element = TOC.create_toc_source(
        title="title",
        outline_level=2,
        title_style="Standard",
        entry_style="Standard_%s",
    )
    assert str(element) == "title"


def test_toc_create_toc_source_bad_args():
    element = TOC.create_toc_source(
        title="title",
        outline_level=2,
        title_style="",
        entry_style="",
    )
    assert str(element) == "title"


def test_toc_get_formatted_text(sample):
    toc = TOC("Table des matières")
    toc.fill(sample)
    result = toc.get_formatted_text()
    assert result.startswith("Table des matières\n1. Level 1")


def test_toc_get_formatted_text_ctx(sample):
    toc = TOC("Table des matières")
    toc.fill(sample)
    ctx = {"rst_mode": True}
    result = toc.get_formatted_text(ctx)
    assert result == "\n.. contents::\n\n"


def test_toc_outline_level():
    toc = TOC("Table des matières")
    result = toc.outline_level
    assert result == 0


def test_toc_outline_level_no_source():
    toc = TOC("Table des matières")
    source = toc.get_element("text:table-of-content-source")
    source.delete()
    result = toc.outline_level
    assert result is None


def test_toc_set_outline_level():
    toc = TOC("Table des matières")
    toc.outline_level = 2
    result = toc.outline_level
    assert result == 2


def test_toc_set_outline_level_no_source():
    toc = TOC("Table des matières")
    source = toc.get_element("text:table-of-content-source")
    source.delete()
    toc.outline_level = 2
    result = toc.outline_level
    assert result == 2


def test_toc_set_some_body():
    toc = TOC("Table des matières")
    toc.body = Element.from_tag("text:index-body")
    assert toc.body.serialize() == "<text:index-body/>"


def test_toc_set_none_body():
    toc = TOC("Table des matières")
    toc.body = None
    assert toc.body.serialize() == "<text:index-body/>"


def test_toc_set_toc_title_1():
    toc = TOC("Table des matières")
    source = toc.get_element("text:table-of-content-source")
    source.delete()
    toc.set_toc_title("new title")
    assert str(toc).startswith("new title")


def test_toc_set_toc_title_2():
    toc = TOC("Table des matières")
    toc.set_toc_title("new title")
    assert str(toc).startswith("new title")


def test_toc_set_toc_title_3():
    toc = TOC("Table des matières")
    toc.set_toc_title("new title", style="Standard")
    assert str(toc).startswith("new title")


def test_toc_set_toc_title_4():
    toc = TOC("Table des matières")
    index_title = toc.body.get_element(IndexTitle._tag)
    index_title.clear()
    toc.set_toc_title("new title", style="Standard")
    assert str(toc).startswith("new title")


def test_toc_set_toc_title_5():
    toc = TOC("Table des matières")
    index_title = toc.body.get_element(IndexTitle._tag)
    index_title.clear()
    toc.set_toc_title(
        "new title",
        style="Standard",
        text_style="Standard",
    )
    assert str(toc).startswith("new title")


def test_toc_get_title_1():
    toc = TOC("Table des matières")
    result = toc.get_title()
    assert result == "Table des matières"


def test_toc_get_title_2():
    toc = TOC("Table des matières")
    toc.body.delete()
    result = toc.get_title()
    assert result == ""


def test_toc_get_title_3():
    toc = TOC("Table des matières")
    index_title = toc.body.get_element(IndexTitle._tag)
    index_title.delete()
    result = toc.get_title()
    assert result == ""
